from __future__ import annotations

from dataclasses import dataclass
from io import BufferedIOBase
from pathlib import Path
from typing import Sequence

import numpy as np
import numpy.typing as npt

import doppy
from doppy import options
from doppy.product.stare import PulsesPerRay, RayAccumulationTime, Stare


@dataclass
class StareDepol:
    """
    Stare product with depolarisation ratio derived from co-polarised and
    cross-polarised stare data.

    Attributes:
    -----------
    time
        An array of datetime64 objects representing the observation times.
    radial_distance
        An array of radial distances from the observation point, in meters.
    elevation
        An array of elevation angles corresponding to the observation points, in
        degrees.
    beta
        An array of backscatter coefficients for the co-polarised signal, in
        sr-1 m-1.
    beta_cross
        An array of backscatter coefficients for the cross-polarised signal, in
        sr-1 m-1.
    radial_velocity
        An array of radial velocities of the co-polarised signal, in m s-1.
    mask_beta
        A boolean array indicating signal (True) or noise (False) data points.
    mask_radial_velocity
        A boolean array indicating signal (True) or noise (False) data points.
    depolarisation
        An array of depolarisation ratios calculated as the ratio of
        co-polarised to cross-polarised backscatter coefficients.
    wavelength
        The wavelength of the lidar, in meters.
    system_id
        A string identifier for the lidar.


    Raises
    ------
    ValueError
        If the input `co` and `cross` products have mismatched wavelengths,
        system IDs, radial distances, or elevation angles, this exception is
        raised.


    References
    ----------
    Aerosol particle depolarization ratio at 1565 nm measured with a Halo Doppler lidar
        authors: Ville Vakkari, Holger Baars, Stephanie Bohlmann, Johannes Bühl,
            Mika Komppula, Rodanthi-Elisavet Mamouri, and Ewan James O'Connor
        doi: https://doi.org/10.5194/acp-21-5807-2021
    """

    time: npt.NDArray[np.datetime64]
    radial_distance: npt.NDArray[np.float64]
    elevation: npt.NDArray[np.float64]
    beta: npt.NDArray[np.float64]
    beta_cross: npt.NDArray[np.float64]
    radial_velocity: npt.NDArray[np.float64]
    mask_beta: npt.NDArray[np.bool_]
    mask_radial_velocity: npt.NDArray[np.bool_]
    depolarisation: npt.NDArray[np.float64]
    polariser_bleed_through: float
    wavelength: float
    system_id: str
    ray_info: RayAccumulationTime | PulsesPerRay

    def __init__(
        self,
        co: Stare,
        cross: Stare,
        polariser_bleed_through: float = 0.0,
    ):
        """
        Parameters
        ----------
        co: Stare
            The co-polarised data.
        cross: Stare
            The cross-polarised data. The `cross.time` array is expected to be sorted.
        polariser_bleed_through: float, default=0.0
            The amount of bleed-through from the polariser.
        """

        if co.beta.shape[1] != cross.beta.shape[1]:
            raise doppy.exceptions.ShapeError(
                "Range dimension mismatch in co and cross: "
                f"{co.beta.shape[1]} vs {cross.beta.shape[1]}"
            )

        if not np.isclose(co.wavelength, cross.wavelength):
            raise ValueError(
                "Different wavelength in co and cross: "
                f"{co.wavelength} vs {cross.wavelength}"
            )
        if co.system_id != cross.system_id:
            raise ValueError(
                "Different system ID in co and cross: "
                f"{co.system_id} vs {cross.system_id}"
            )
        if not np.allclose(co.radial_distance, cross.radial_distance, atol=1):
            raise ValueError("Different radial distance in co and cross")

        ind = np.searchsorted(cross.time, co.time, side="left")
        pick_ind = ind < len(cross.time)
        time_diff_threshold = 2 * np.median(np.diff(co.time))
        co_cross_timediff_below_threshold = (
            cross.time[ind[pick_ind]] - co.time[pick_ind] < time_diff_threshold
        )
        pick_ind[pick_ind] &= co_cross_timediff_below_threshold

        if not np.allclose(
            co.elevation[pick_ind], cross.elevation[ind[pick_ind]], atol=1
        ):
            raise ValueError("Different elevation in co and cross")

        depolarisation = np.full_like(co.beta, np.nan)
        co_beta = co.beta[pick_ind]
        depolarisation[pick_ind] = (
            cross.beta[ind[pick_ind]] - polariser_bleed_through * co_beta
        ) / co_beta
        cross_beta = np.full_like(co.beta, np.nan)
        cross_beta[pick_ind] = cross.beta[ind[pick_ind]]

        self.time = co.time
        self.radial_distance = co.radial_distance
        self.elevation = co.elevation
        self.beta = co.beta
        self.beta_cross = cross_beta
        self.radial_velocity = co.radial_velocity
        self.mask_beta = co.mask_beta
        self.mask_radial_velocity = co.mask_radial_velocity
        self.depolarisation = depolarisation
        self.polariser_bleed_through = polariser_bleed_through
        self.wavelength = co.wavelength
        self.system_id = co.system_id
        self.ray_info = co.ray_info

    @property
    def mask_depolarisation(self) -> npt.NDArray[np.bool_]:
        return np.isnan(self.depolarisation)

    @property
    def mask_beta_cross(self) -> npt.NDArray[np.bool_]:
        return np.isnan(self.beta_cross)

    @classmethod
    def from_halo_data(
        cls,
        co_data: Sequence[str]
        | Sequence[Path]
        | Sequence[bytes]
        | Sequence[BufferedIOBase],
        co_data_bg: Sequence[str]
        | Sequence[Path]
        | Sequence[tuple[bytes, str]]
        | Sequence[tuple[BufferedIOBase, str]],
        cross_data: Sequence[str]
        | Sequence[Path]
        | Sequence[bytes]
        | Sequence[BufferedIOBase],
        cross_data_bg: Sequence[str]
        | Sequence[Path]
        | Sequence[tuple[bytes, str]]
        | Sequence[tuple[BufferedIOBase, str]],
        bg_correction_method: options.BgCorrectionMethod,
        polariser_bleed_through: float = 0,
    ) -> StareDepol:
        co = Stare.from_halo_data(
            data=co_data, data_bg=co_data_bg, bg_correction_method=bg_correction_method
        )
        cross = Stare.from_halo_data(
            data=cross_data,
            data_bg=cross_data_bg,
            bg_correction_method=bg_correction_method,
        )
        return cls(co, cross, polariser_bleed_through)

    def write_to_netcdf(self, filename: str | Path) -> None:
        with doppy.netcdf.Dataset(filename) as nc:
            nc.add_dimension("time")
            nc.add_dimension("range")
            nc.add_time(
                name="time",
                dimensions=("time",),
                standard_name="time",
                long_name="Time UTC",
                data=self.time,
                dtype="f8",
            )
            nc.add_variable(
                name="range",
                dimensions=("range",),
                units="m",
                data=self.radial_distance,
                dtype="f4",
            )
            nc.add_variable(
                name="elevation",
                dimensions=("time",),
                units="degrees",
                data=self.elevation,
                dtype="f4",
                long_name="elevation from horizontal",
            )
            nc.add_variable(
                name="beta_raw",
                dimensions=("time", "range"),
                units="sr-1 m-1",
                data=self.beta,
                dtype="f4",
            )
            nc.add_variable(
                name="beta",
                dimensions=("time", "range"),
                units="sr-1 m-1",
                data=self.beta,
                dtype="f4",
                mask=self.mask_beta,
            )
            nc.add_variable(
                name="v",
                dimensions=("time", "range"),
                units="m s-1",
                long_name="Doppler velocity",
                data=self.radial_velocity,
                dtype="f4",
                mask=self.mask_radial_velocity,
            )
            nc.add_scalar_variable(
                name="wavelength",
                units="m",
                standard_name="radiation_wavelength",
                data=self.wavelength,
                dtype="f4",
            )
            nc.add_variable(
                name="depolarisation_raw",
                dimensions=("time", "range"),
                units="1",
                data=self.depolarisation,
                dtype="f4",
                mask=self.mask_depolarisation,
            )
            nc.add_variable(
                name="depolarisation",
                dimensions=("time", "range"),
                units="1",
                data=self.depolarisation,
                dtype="f4",
                mask=self.mask_beta | self.mask_depolarisation,
            )
            nc.add_variable(
                name="beta_cross_raw",
                dimensions=("time", "range"),
                units="sr-1 m-1",
                data=self.beta_cross,
                mask=self.mask_beta_cross,
                dtype="f4",
            )
            nc.add_variable(
                name="beta_cross",
                dimensions=("time", "range"),
                units="sr-1 m-1",
                data=self.beta_cross,
                mask=self.mask_beta | self.mask_beta_cross,
                dtype="f4",
            )
            nc.add_scalar_variable(
                name="polariser_bleed_through",
                units="1",
                long_name="Polariser bleed-through",
                data=self.polariser_bleed_through,
                dtype="f4",
            )
            match self.ray_info:
                case RayAccumulationTime(value):
                    nc.add_scalar_variable(
                        name="ray_accumulation_time",
                        units="s",
                        long_name="ray accumulation time",
                        data=value,
                        dtype="f4",
                    )
                case PulsesPerRay(value):
                    nc.add_scalar_variable(
                        name="pulses_per_ray",
                        units="1",
                        long_name="pulses per ray",
                        data=value,
                        dtype="u4",
                    )
            nc.add_attribute("serial_number", self.system_id)
            nc.add_attribute("doppy_version", doppy.__version__)



DEFAULT_BEAM_ENERGY = 1e-5
DEFAULT_EFFECTIVE_DIAMETER = 25e-3


class Halo:
    wavelength = 1.565e-6  # [m]
    receiver_bandwidth = 50e6  # [Hz]
    beam_energy = DEFAULT_BEAM_ENERGY
    effective_diameter = DEFAULT_EFFECTIVE_DIAMETER


class WindCube:
    # https://doi.org/10.5194/essd-13-3539-2021
    wavelength = 1.54e-6  # [m]
    receiver_bandwidth = 55e6  # [Hz]
    beam_energy = DEFAULT_BEAM_ENERGY
    effective_diameter = 50e-3  # [m]
    focus = 1e3  # [m]



from doppy.product.stare import Stare
from doppy.product.stare_depol import StareDepol
from doppy.product.wind import Options as WindOptions
from doppy.product.wind import Wind

__all__ = ["Stare", "StareDepol", "Wind", "WindOptions"]



import gzip
import io

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from doppy.data import exceptions
from doppy.data.cache import cached_record


class Api:
    def __init__(self, cache: bool = False) -> None:
        retries = Retry(total=10, backoff_factor=0.2)
        adapter = HTTPAdapter(max_retries=retries)
        session = requests.Session()
        session.mount("https://", adapter)
        session.mount("http://", adapter)
        self.session = session
        self.api_endpoint = "https://cloudnet.fmi.fi/api"
        self.cache = cache

    def get(self, path: str, params: dict[str, str | list[str]]) -> list:
        res = self.session.get(
            f"{self.api_endpoint}/{path}", params=params, timeout=1800
        )
        if res.ok:
            data = res.json()
            if isinstance(data, list):
                return data
            raise exceptions.ApiRequestError(
                f"Unexpected response type from api: {type(data)}"
            )
        raise exceptions.ApiRequestError(f"Api request error: {res.status_code}")

    def get_raw_records(self, site: str, date: str) -> list:
        return self.get(
            "raw-files",
            params={
                "instrument": [
                    "halo-doppler-lidar",
                    "wls100s",
                    "wls200s",
                    "wls400s",
                    "wls70",
                ],
                "site": site,
                "date": date,
            },
        )

    def get_record_content(self, rec: dict) -> io.BytesIO:
        if self.cache:
            return cached_record(rec, self.session)
        content = self.session.get(rec["downloadUrl"]).content
        if rec["filename"].endswith(".gz"):
            content = gzip.decompress(content)
        return io.BytesIO(content)



from __future__ import annotations

import functools
import logging
import re
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from io import BufferedIOBase
from os.path import commonprefix
from pathlib import Path
from typing import Any, Sequence, cast

import numpy as np
import numpy.typing as npt
from numpy import datetime64, timedelta64

import doppy
from doppy import exceptions
from doppy.raw.utils import bytes_from_src
from doppy.utils import merge_all_equal


@dataclass
class HaloHpl:
    header: HaloHplHeader
    time: npt.NDArray[datetime64]  # dim: (time, )
    radial_distance: npt.NDArray[np.float64]  # dim: (radial_distance, )
    azimuth: npt.NDArray[np.float64]  # dim: (time, )
    elevation: npt.NDArray[np.float64]  # dim: (time, )
    pitch: npt.NDArray[np.float64] | None  # dim: (time, )
    roll: npt.NDArray[np.float64] | None  # dim: (time, )
    radial_velocity: npt.NDArray[np.float64]  # dim: (time, radial_distance)
    intensity: npt.NDArray[np.float64]  # dim: (time, radial_distance)
    beta: npt.NDArray[np.float64]  # dim: (time, radial_distance)
    spectral_width: npt.NDArray[np.float64] | None  # dim: (time, radial_distance )

    @classmethod
    def from_srcs(
        cls, data: Sequence[str | bytes | Path | BufferedIOBase]
    ) -> list[HaloHpl]:
        data_bytes = [bytes_from_src(src) for src in data]
        raw_dicts = doppy.rs.raw.halo_hpl.from_bytes_srcs(data_bytes)
        raws = []
        for r in raw_dicts:
            try:
                raws.append(_raw_tuple2halo_hpl(r))
            except exceptions.RawParsingError as err:
                logging.warning("Skipping %s: %s", r[0].get("filename"), err)
        return raws

    @classmethod
    def from_src(cls, data: str | Path | bytes | BufferedIOBase) -> HaloHpl:
        data_bytes = bytes_from_src(data)
        try:
            return _raw_tuple2halo_hpl(doppy.rs.raw.halo_hpl.from_bytes_src(data_bytes))
        except RuntimeError as err:
            raise exceptions.RawParsingError(err) from err

    def __getitem__(
        self,
        index: int
        | slice
        | list[int]
        | npt.NDArray[np.int64]
        | npt.NDArray[np.bool_]
        | tuple[slice, slice],
    ) -> HaloHpl:
        if isinstance(index, (int, slice, list, np.ndarray)):
            return HaloHpl(
                header=self.header,
                time=self.time[index],
                radial_distance=self.radial_distance,
                azimuth=self.azimuth[index],
                elevation=self.elevation[index],
                radial_velocity=self.radial_velocity[index],
                intensity=self.intensity[index],
                beta=self.beta[index],
                pitch=self.pitch[index] if self.pitch is not None else None,
                roll=self.roll[index] if self.roll is not None else None,
                spectral_width=self.spectral_width[index]
                if self.spectral_width is not None
                else None,
            )
        raise TypeError

    @classmethod
    def merge(cls, raws: Sequence[HaloHpl]) -> HaloHpl:
        return cls(
            header=_merge_headers([r.header for r in raws]),
            time=np.concatenate(tuple(r.time for r in raws)),
            radial_distance=raws[0].radial_distance,
            azimuth=np.concatenate(tuple(r.azimuth for r in raws)),
            elevation=np.concatenate(tuple(r.elevation for r in raws)),
            radial_velocity=np.concatenate(tuple(r.radial_velocity for r in raws)),
            intensity=np.concatenate(tuple(r.intensity for r in raws)),
            beta=np.concatenate(tuple(r.beta for r in raws)),
            pitch=_merge_float_arrays_or_nones(tuple(r.pitch for r in raws)),
            roll=_merge_float_arrays_or_nones(tuple(r.roll for r in raws)),
            spectral_width=_merge_float_arrays_or_nones(
                tuple(r.spectral_width for r in raws)
            ),
        )

    @functools.cached_property
    def azimuth_angles(self) -> set[int]:
        return set(int(x) % 360 for x in np.round(self.azimuth))

    @functools.cached_property
    def elevation_angles(self) -> set[int]:
        return set(int(x) for x in np.round(self.elevation))

    @functools.cached_property
    def time_diffs(self) -> set[int]:
        return set(np.diff(self.time.astype("datetime64[s]").astype("int")))

    @functools.cached_property
    def median_time_diff(self) -> float:
        med = np.round(
            np.median(
                np.diff(1e-6 * self.time.astype("datetime64[us]").astype("float"))
            ),
            2,
        )
        if isinstance(med, float):
            return med
        raise TypeError

    def sorted_by_time(self) -> HaloHpl:
        sort_indices = np.argsort(self.time)
        return self[sort_indices]

    def non_strictly_increasing_timesteps_removed(self) -> HaloHpl:
        if len(self.time) == 0:
            return self
        mask = np.ones_like(self.time, dtype=np.bool_)
        latest_time = self.time[0]
        for i, t in enumerate(self.time[1:], start=1):
            if t <= latest_time:
                mask[i] = False
            else:
                latest_time = t
        return self[mask]

    def nans_removed(self) -> HaloHpl:
        is_ok = np.array(~np.isnan(self.intensity).any(axis=1), dtype=np.bool_)
        return self[is_ok]


@dataclass(slots=True)
class HaloHplHeader:
    filename: str
    gate_points: int
    nrays: int | None
    nwaypoints: int | None
    ngates: int
    pulses_per_ray: int
    range_gate_length: float
    resolution: float
    scan_type: str
    focus_range: int
    start_time: datetime64
    system_id: str
    instrument_spectral_width: float | None

    def mergeable_hash(self) -> int:
        return hash(
            (
                self.gate_points,
                self.nrays,
                self.nwaypoints,
                self.ngates,
                self.pulses_per_ray,
                round(self.range_gate_length, 1),
                round(self.resolution, 1),
                self.scan_type,
                self.focus_range,
                self.system_id,
                round(x, 1)
                if isinstance((x := self.instrument_spectral_width), float)
                else None,
            )
        )

    @classmethod
    def from_dict(cls, data: dict[bytes, bytes]) -> HaloHplHeader:
        return cls(
            filename=data[b"Filename"].decode(),
            gate_points=int(data[b"Gate length (pts)"]),
            nrays=(
                int(data[b"No. of rays in file"])
                if b"No. of rays in file" in data
                else None
            ),
            nwaypoints=(
                int(data[b"No. of waypoints in file"])
                if b"No. of waypoints in file" in data
                else None
            ),
            ngates=int(data[b"Number of gates"]),
            pulses_per_ray=int(data[b"Pulses/ray"]),
            range_gate_length=float(data[b"Range gate length (m)"]),
            resolution=float(data[b"Resolution (m/s)"]),
            scan_type=data[b"Scan type"].decode(),
            focus_range=int(data[b"Focus range"]),
            start_time=_parser_start_time(data[b"Start time"]),
            system_id=data[b"System ID"].decode(),
            instrument_spectral_width=(
                float(data[b"instrument_spectral_width"])
                if b"instrument_spectral_width" in data
                else None
            ),
        )


def _merge_headers(headers: list[HaloHplHeader]) -> HaloHplHeader:
    return HaloHplHeader(
        filename=commonprefix([h.filename for h in headers]),
        start_time=np.min([h.start_time for h in headers]),
        **{
            key: merge_all_equal(key, [getattr(h, key) for h in headers])
            for key in (
                "gate_points",
                "nrays",
                "nwaypoints",
                "ngates",
                "pulses_per_ray",
                "range_gate_length",
                "resolution",
                "scan_type",
                "focus_range",
                "system_id",
                "instrument_spectral_width",
            )
        },
    )


def _merge_float_arrays_or_nones(
    arrs: tuple[npt.NDArray[np.float64] | None, ...],
) -> npt.NDArray[np.float64] | None:
    isnone = tuple(x is None for x in arrs)
    if all(isnone):
        return None
    if any(isnone):
        raise ValueError
    arrs = cast(tuple[npt.NDArray[np.float64], ...], arrs)
    return np.concatenate(arrs, axis=0)


def _raw_tuple2halo_hpl(
    raw_tuple: tuple[dict[str, Any], dict[str, npt.NDArray[np.float64] | None]],
) -> HaloHpl:
    header_dict, data_dict = raw_tuple
    header = HaloHplHeader(
        filename=str(header_dict["filename"]),
        gate_points=int(header_dict["gate_points"]),
        nrays=int(header_dict["nrays"]) if header_dict["nrays"] is not None else None,
        nwaypoints=int(header_dict["nwaypoints"])
        if header_dict["nwaypoints"] is not None
        else None,
        ngates=int(header_dict["ngates"]),
        pulses_per_ray=int(header_dict["pulses_per_ray"]),
        range_gate_length=float(header_dict["range_gate_length"]),
        resolution=float(header_dict["resolution"]),
        scan_type=str(header_dict["scan_type"]),
        focus_range=int(header_dict["focus_range"]),
        start_time=datetime64(
            datetime.fromtimestamp(header_dict["start_time"], timezone.utc).replace(
                tzinfo=None
            )
        ),
        system_id=str(header_dict["system_id"]),
        instrument_spectral_width=float(header_dict["instrument_spectral_width"])
        if header_dict["instrument_spectral_width"] is not None
        else None,
    )
    expected_range = np.arange(header.ngates, dtype=np.float64)

    if any(
        data_dict[key] is None
        for key in (
            "range",
            "time",
            "radial_distance",
            "azimuth",
            "elevation",
            "radial_velocity",
            "intensity",
            "beta",
        )
    ):
        raise TypeError
    range_ = cast(npt.NDArray[np.float64], data_dict["range"]).reshape(
        -1, header.ngates
    )
    radial_distance = cast(npt.NDArray[np.float64], data_dict["radial_distance"])
    azimuth = cast(npt.NDArray[np.float64], data_dict["azimuth"])
    elevation = cast(npt.NDArray[np.float64], data_dict["elevation"])
    radial_velocity = cast(
        npt.NDArray[np.float64], data_dict["radial_velocity"]
    ).reshape(-1, header.ngates)
    intensity = cast(npt.NDArray[np.float64], data_dict["intensity"]).reshape(
        -1, header.ngates
    )
    beta = cast(npt.NDArray[np.float64], data_dict["beta"]).reshape(-1, header.ngates)
    if not np.isclose(range_, expected_range).all():
        raise exceptions.RawParsingError(
            "Incoherent range gates: Number of gates in the middle of the file"
        )
    return HaloHpl(
        header=header,
        time=_convert_time(
            header.start_time, cast(npt.NDArray[np.float64], data_dict["time"])
        ),
        radial_distance=radial_distance,
        azimuth=azimuth,
        elevation=elevation,
        pitch=data_dict["pitch"] if data_dict["pitch"] is not None else None,
        roll=data_dict["roll"] if data_dict["roll"] is not None else None,
        radial_velocity=radial_velocity,
        intensity=intensity,
        beta=beta,
        spectral_width=data_dict["spectral_width"].reshape(-1, header.ngates)
        if data_dict["spectral_width"] is not None
        else None,
    )


def _convert_time(
    start_time: datetime64, decimal_time: npt.NDArray[np.float64]
) -> npt.NDArray[datetime64]:
    """
    Parameters
    ----------
    start_time: unix-time
    decimal_time: hours since beginning of the day of start_time
    """
    HOURS_TO_MICROSECONDS = 3600000000.0
    start_of_day = start_time.astype("datetime64[D]").astype("datetime64[us]")
    delta_hours = (decimal_time * HOURS_TO_MICROSECONDS).astype("timedelta64[us]")
    return np.array(start_of_day + delta_hours, dtype=datetime64)


def _parser_start_time(s: bytes) -> datetime64:
    return datetime64(datetime.strptime(s.decode(), "%Y%m%d %H:%M:%S.%f"))


def _from_src(data: BufferedIOBase) -> HaloHpl:
    head = data.read(1000)
    match_header_div = re.search(b"\\*\\*\\*\\*.*\n+", head, re.MULTILINE)
    if match_header_div is None:
        raise exceptions.RawParsingError("Cannot find header divider '****'")
    data.seek(0)
    _, div = match_header_div.span()
    header_bytes = data.read(div)
    header = _read_header(header_bytes)
    data_bytes = data.read()
    res = _read_data(data_bytes, header)
    return res


def _read_header(data: bytes) -> HaloHplHeader:
    data = data.strip()
    data_dict = {}
    expected_header_rows = [
        b"Altitude of measurement (center of gate) = (range gate + 0.5) * Gate length",
        b"Range of measurement (center of gate) = (range gate + 0.5) * Gate length",
        b"Data line 1: Decimal time (hours)  Azimuth (degrees)  Elevation (degrees) "
        b"Pitch (degrees) Roll (degrees)",
        b"Data line 1: Decimal time (hours)  Azimuth (degrees)  Elevation (degrees)",
        b"f9.6,1x,f6.2,1x,f6.2",
        b"Data line 2: Range Gate  Doppler (m/s)  Intensity (SNR + 1)  Beta (m-1 sr-1)",
        b"Data line 2: Range Gate  Doppler (m/s)  Intensity (SNR + 1)  Beta (m-1 sr-1) "
        b"Spectral Width",
        b"i3,1x,f6.4,1x,f8.6,1x,e12.6 - repeat for no. gates",
        b"i3,1x,f6.4,1x,f8.6,1x,e12.6,1x,f6.4 - repeat for no. gates",
        b"****",
    ]
    for line in data.split(b"\r\n"):
        split = line.split(b":\t")
        if len(split) == 2:
            key, val = split
            data_dict[key] = val
        else:
            (val,) = split
            if m := re.match(rb"\*\*\*\* Instrument spectral width = (.*)", val):
                data_dict[b"instrument_spectral_width"] = m.group(1)
            elif val not in expected_header_rows:
                raise ValueError(f"Unexpected row '{val!r}'")
    return HaloHplHeader.from_dict(data_dict)


def _read_data(data: bytes, header: HaloHplHeader) -> HaloHpl:
    if not data:
        raise exceptions.RawParsingError("No data found")
    data = data.strip()
    data = data.replace(
        b"\x00", b""
    )  # Some files contain null characters between profiles
    data_lines = data.split(b"\r\n")

    i = 0
    while i + 1 < len(data_lines) and data_lines[i + 1].strip().split()[0] != b"0":
        i += 1
    del data_lines[:i]

    i = len(data_lines) - 1
    while (
        i - 1 >= 0
        and header.ngates > 1
        and len(data_lines[i].strip().split()) != len(data_lines[i - 1].strip().split())
    ):
        i -= 1
    del data_lines[i + 1 :]

    trailing_lines = len(data_lines) % (header.ngates + 1)
    if trailing_lines > 0:
        del data_lines[-trailing_lines:]

    data1D_lines = data_lines[:: header.ngates + 1]
    data1D = [list(map(float, line.split())) for line in data1D_lines]
    try:
        data1Darr = np.array(data1D)
    except ValueError as err:
        if "inhomogeneous" in str(err):
            raise exceptions.RawParsingError(
                "Inhomogeneous raw data. "
                "Probable reason: Number of gates changes in middle of the file"
            ) from err
        else:
            raise

    del data_lines[:: header.ngates + 1]
    data2D = [list(map(float, line.split())) for line in data_lines]
    data2Darr = np.array(data2D)

    decimal_time = data1Darr[:, 0]
    time = header.start_time.astype("datetime64[D]") + np.array(
        list(map(_decimal_time2timedelta, decimal_time))
    )
    azimuth = data1Darr[:, 1]
    elevation = data1Darr[:, 2]
    pitch = data1Darr[:, 3] if data1Darr.shape[1] > 3 else None
    roll = data1Darr[:, 4] if data1Darr.shape[1] > 4 else None

    ntimes = len(decimal_time)

    data2Darr_reshape = data2Darr.reshape(ntimes, header.ngates, -1)

    gate = data2Darr_reshape[:, :, 0]
    gate_expected = np.arange(len(gate[0])).astype("float64")
    if not all(np.allclose(gate_expected, gate[i, :]) for i in range(gate.shape[0])):
        raise ValueError("all gate indices should be equal")
    radial_distance = (gate_expected + 0.5) * header.range_gate_length

    radial_velocity = data2Darr_reshape[:, :, 1]
    intensity = data2Darr_reshape[:, :, 2]
    beta = data2Darr_reshape[:, :, 3]

    spectral_width = (
        data2Darr_reshape[:, :, 4] if data2Darr_reshape.shape[2] > 4 else None
    )

    return HaloHpl(
        header,
        time,
        radial_distance,
        azimuth,
        elevation,
        pitch,
        roll,
        radial_velocity,
        intensity,
        beta,
        spectral_width,
    )


def _decimal_time2timedelta(h: float) -> timedelta64:
    return timedelta64(timedelta(hours=h))



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timezone
from io import BufferedIOBase
from pathlib import Path
from typing import Any, Sequence

import numpy as np
import numpy.typing as npt
from numpy import datetime64

import doppy
from doppy import exceptions
from doppy.raw.utils import bytes_from_src
from doppy.utils import merge_all_equal


@dataclass
class Wls77:
    time: npt.NDArray[datetime64]  # dim: (time, )
    altitude: npt.NDArray[np.float64]  # dim: (altitude, )
    position: npt.NDArray[np.float64]  # dim: (time, )
    temperature: npt.NDArray[np.float64]  # dim: (time, )
    wiper_count: npt.NDArray[np.float64]  # dim: (time, )
    cnr: npt.NDArray[np.float64]  # dim: (time, altitude)
    radial_velocity: npt.NDArray[np.float64]  # dim: (time, altitude)
    radial_velocity_deviation: npt.NDArray[np.float64]  # dim: (time, altitude)
    wind_speed: npt.NDArray[np.float64]  # dim: (time, altitude)
    wind_direction: npt.NDArray[np.float64]  # dim: (time, altitude)
    zonal_wind: npt.NDArray[np.float64]  # u := zonal wind?, dim: (time, altitude)
    meridional_wind: npt.NDArray[
        np.float64
    ]  # v := meridional wind?, dim: (time, altitude)
    vertical_wind: npt.NDArray[np.float64]  # w := vertical wind?, dim: (time, altitude)
    cnr_threshold: float
    system_id: str

    @classmethod
    def from_srcs(
        cls, data: Sequence[str | bytes | Path | BufferedIOBase]
    ) -> list[Wls77]:
        data_bytes = [bytes_from_src(src) for src in data]
        raws = doppy.rs.raw.wls77.from_bytes_srcs(data_bytes)
        try:
            return [_raw_rs_to_wls77(r) for r in raws]
        except RuntimeError as err:
            raise exceptions.RawParsingError(err) from err

    @classmethod
    def from_src(cls, data: str | Path | bytes | BufferedIOBase) -> Wls77:
        data_bytes = bytes_from_src(data)
        try:
            return _raw_rs_to_wls77(doppy.rs.raw.wls77.from_bytes_src(data_bytes))
        except RuntimeError as err:
            raise exceptions.RawParsingError(err) from err

    def __getitem__(
        self,
        index: int
        | slice
        | list[int]
        | npt.NDArray[np.int64]
        | npt.NDArray[np.bool_]
        | tuple[slice, slice],
    ) -> Wls77:
        if isinstance(index, (int, slice, list, np.ndarray)):
            return Wls77(
                time=self.time[index],
                altitude=self.altitude,
                position=self.position[index],
                temperature=self.temperature[index],
                wiper_count=self.wiper_count[index],
                cnr=self.cnr[index],
                radial_velocity=self.radial_velocity[index],
                radial_velocity_deviation=self.radial_velocity_deviation[index],
                wind_speed=self.wind_speed[index],
                wind_direction=self.wind_direction[index],
                zonal_wind=self.zonal_wind[index],
                meridional_wind=self.meridional_wind[index],
                vertical_wind=self.vertical_wind[index],
                system_id=self.system_id,
                cnr_threshold=self.cnr_threshold,
            )
        raise TypeError

    def sorted_by_time(self) -> Wls77:
        sort_indices = np.argsort(self.time)
        return self[sort_indices]

    @classmethod
    def merge(cls, raws: Sequence[Wls77]) -> Wls77:
        return cls(
            time=np.concatenate(tuple(r.time for r in raws)),
            altitude=raws[0].altitude,
            position=np.concatenate(tuple(r.position for r in raws)),
            temperature=np.concatenate(tuple(r.temperature for r in raws)),
            wiper_count=np.concatenate(tuple(r.wiper_count for r in raws)),
            cnr=np.concatenate(tuple(r.cnr for r in raws)),
            radial_velocity=np.concatenate(tuple(r.radial_velocity for r in raws)),
            radial_velocity_deviation=np.concatenate(
                tuple(r.radial_velocity_deviation for r in raws)
            ),
            wind_speed=np.concatenate(tuple(r.wind_speed for r in raws)),
            wind_direction=np.concatenate(tuple(r.wind_direction for r in raws)),
            zonal_wind=np.concatenate(tuple(r.zonal_wind for r in raws)),
            meridional_wind=np.concatenate(tuple(r.meridional_wind for r in raws)),
            vertical_wind=np.concatenate(tuple(r.vertical_wind for r in raws)),
            system_id=merge_all_equal("system_id", [r.system_id for r in raws]),
            cnr_threshold=merge_all_equal(
                "cnr_threshold", [r.cnr_threshold for r in raws]
            ),
        )

    def non_strictly_increasing_timesteps_removed(self) -> Wls77:
        if len(self.time) == 0:
            return self
        mask = np.ones_like(self.time, dtype=np.bool_)
        latest_time = self.time[0]
        for i, t in enumerate(self.time[1:], start=1):
            if t <= latest_time:
                mask[i] = False
            else:
                latest_time = t
        return self[mask]


def _raw_rs_to_wls77(
    raw: dict[str, Any],
) -> Wls77:
    time_ts = raw["time"]
    time = np.array(
        [
            datetime64(datetime.fromtimestamp(ts, timezone.utc).replace(tzinfo=None))
            for ts in time_ts
        ]
    )

    n = time.size

    return Wls77(
        time=time,
        altitude=np.array(raw["altitude"], dtype=np.float64),
        position=np.array(raw["position"], dtype=np.float64),
        temperature=np.array(raw["temperature"], dtype=np.float64),
        wiper_count=np.array(raw["wiper_count"], dtype=np.float64),
        cnr=np.array(raw["cnr"], dtype=np.float64).reshape(n, -1),
        radial_velocity=np.array(raw["radial_velocity"], dtype=np.float64).reshape(
            n, -1
        ),
        radial_velocity_deviation=np.array(
            raw["radial_velocity_deviation"], dtype=np.float64
        ).reshape(n, -1),
        wind_speed=np.array(raw["wind_speed"], dtype=np.float64).reshape(n, -1),
        wind_direction=np.array(raw["wind_direction"], dtype=np.float64).reshape(n, -1),
        zonal_wind=np.array(raw["zonal_wind"], dtype=np.float64).reshape(n, -1),
        meridional_wind=np.array(raw["meridional_wind"], dtype=np.float64).reshape(
            n, -1
        ),
        vertical_wind=np.array(raw["vertical_wind"], dtype=np.float64).reshape(n, -1),
        cnr_threshold=raw["cnr_threshold"],
        system_id=raw["system_id"],
    )



from .halo_bg import HaloBg
from .halo_hpl import HaloHpl
from .halo_sys_params import HaloSysParams
from .windcube import WindCube, WindCubeFixed
from .wls70 import Wls70
from .wls77 import Wls77

__all__ = [
    "HaloHpl",
    "HaloBg",
    "HaloSysParams",
    "WindCube",
    "WindCubeFixed",
    "Wls70",
    "Wls77",
]



import warnings

import numpy as np
import numpy.typing as npt
import scipy


def detect_wind_noise(
    w: npt.NDArray[np.float64],
    height: npt.NDArray[np.float64],
    mask: npt.NDArray[np.bool_],
    window: float = 150,
    stride: int = 1,
) -> npt.NDArray[np.bool_]:
    """
    Parameters
    ----------
    w
        vertical velocity, dims: (time,height)

    height
        dims: (height,)

    mask
        old mask that still contains noisy velocity data,
        mask[t,h] = True iff w[t,h] is noise

    window
        size of window used to compute rolling median in meters

    stride
        stride used to compute rolling median

    Returns
    -------
    improved noise mask such that new_mask[t,h] = True iff w[t,h] is noise
    """
    warnings.simplefilter("ignore", RuntimeWarning)
    v = _rolling_median_over_range(
        height,
        w,
        mask,
        window=window,  # meters
        stride=stride,
        fill_gaps=True,
    )

    th = 2
    diff = np.abs(v - w)
    new_mask = (diff > th) | mask
    new_mask = _remove_one_hot(new_mask)
    return np.array(new_mask, dtype=np.bool_)


def _remove_one_hot(m: npt.NDArray[np.bool_]) -> npt.NDArray[np.bool_]:
    if m.ndim != 2:
        raise ValueError
    if m.shape[1] < 3:
        return m
    x = ~m
    y = np.full(x.shape, np.False_)
    y[:, 0] = x[:, 0] & x[:, 1]
    y[:, 1:-1] = x[:, 1:-1] & (x[:, 2:] | x[:, :-2])
    y[:, -1] = x[:, -1] & x[:, -2]
    return np.array(~y, dtype=np.bool_)


def _rolling_median_over_range(
    range_: npt.NDArray[np.float64],
    arr: npt.NDArray[np.float64],
    mask: npt.NDArray[np.bool_],
    window: float,
    stride: int = 1,
    fill_gaps: bool = False,
) -> npt.NDArray[np.float64]:
    """
    window
        range window in meters
    """
    X = arr.T.copy()
    X[mask.T] = np.nan

    half_window = window / 2

    i = 0
    j = 0
    n = len(range_)
    med = np.full(X.shape, np.nan, dtype=np.float64)
    for k in range(0, n, stride):
        r = range_[k]
        while i + 1 < n and r - range_[i + 1] >= half_window:
            i += 1
        while j + 1 < n and range_[j] - r < half_window:
            j += 1
        if i > k or j < k:
            raise ValueError
        med[k] = np.nanmedian(X[i : j + 1], axis=0)

    if stride != 1 and fill_gaps:
        ind = list(range(0, n, stride))
        f_interp = scipy.interpolate.interp1d(
            range_[ind], med[ind], axis=0, fill_value="extrapolate"
        )
        med_all = f_interp(range_)
        return np.array(med_all.T.copy(), dtype=np.float64)
    return np.array(med.T.copy(), dtype=np.float64)



from __future__ import annotations

import pathlib
import warnings
from types import TracebackType
from typing import Literal, TypeAlias

import netCDF4
import numpy as np
import numpy.typing as npt

NetCDFDataType: TypeAlias = Literal["f4", "f8", "i4", "i8", "u4", "u8"]


class Dataset:
    def __init__(
        self,
        filename: str | pathlib.Path,
        format: Literal["NETCDF4", "NETCDF4_CLASSIC"] = "NETCDF4",
    ) -> None:
        self.nc = netCDF4.Dataset(filename, mode="w", format=format)

    def __enter__(self) -> Dataset:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.close()

    def add_dimension(self, dim: str, size: int | None = None) -> Dataset:
        self.nc.createDimension(dim, size)
        return self

    def add_attribute(self, key: str, val: str) -> Dataset:
        setattr(self.nc, key, val)
        return self

    def add_atribute(self, key: str, val: str) -> Dataset:
        warnings.warn("Use add_attribute", DeprecationWarning, stacklevel=2)
        return self.add_attribute(key, val)

    def add_time(
        self,
        name: str,
        dimensions: tuple[str, ...],
        data: npt.NDArray[np.datetime64],
        dtype: NetCDFDataType,
        standard_name: str | None = None,
        long_name: str | None = None,
    ) -> Dataset:
        time, units, calendar = _convert_time(data)
        var = self.nc.createVariable(name, dtype, dimensions, compression="zlib")
        var.units = units
        var.calendar = calendar
        var.axis = "T"
        var[:] = time
        if standard_name is not None:
            var.standard_name = standard_name
        if long_name is not None:
            var.long_name = long_name
        return self

    def add_variable(
        self,
        name: str,
        dimensions: tuple[str, ...],
        units: str,
        data: npt.NDArray[np.float64],
        dtype: NetCDFDataType,
        standard_name: str | None = None,
        long_name: str | None = None,
        mask: npt.NDArray[np.bool_] | None = None,
    ) -> Dataset:
        fill_value = netCDF4.default_fillvals[dtype] if mask is not None else None
        var = self.nc.createVariable(
            name, dtype, dimensions, fill_value=fill_value, compression="zlib"
        )
        var.units = units
        if mask is not None:
            var[:] = np.ma.masked_array(data, mask)  # type: ignore
        else:
            var[:] = data
        if standard_name is not None:
            var.standard_name = standard_name
        if long_name is not None:
            var.long_name = long_name
        return self

    def add_scalar_variable(
        self,
        name: str,
        units: str,
        data: np.float64 | np.int64 | float | int,
        dtype: NetCDFDataType,
        standard_name: str | None = None,
        long_name: str | None = None,
        mask: npt.NDArray[np.bool_] | None = None,
    ) -> Dataset:
        var = self.nc.createVariable(name, dtype)
        var.units = units
        var[:] = data
        if standard_name is not None:
            var.standard_name = standard_name
        if long_name is not None:
            var.long_name = long_name
        return self

    def close(self) -> None:
        self.nc.close()


def _convert_time(
    time: npt.NDArray[np.datetime64],
) -> tuple[npt.NDArray[np.float64], str, str]:
    """
    Parameters
    ----------
    time : npt.NDArray[np.datetime64["us"]]
        Must be represented in UTC
    """
    if time.dtype != "<M8[us]":
        raise TypeError("time must be datetime64[us]")
    MICROSECONDS_TO_HOURS = 1 / (1e6 * 3600)
    start_of_day = time.min().astype("datetime64[D]")
    hours_since_start_of_day = (time - start_of_day).astype(
        np.float64
    ) * MICROSECONDS_TO_HOURS
    units = f"hours since {np.datetime_as_string(start_of_day)} 00:00:00 +00:00"
    calendar = "standard"
    return hours_since_start_of_day, units, calendar



from __future__ import annotations

import functools
from collections import Counter
from dataclasses import dataclass
from io import BufferedIOBase
from pathlib import Path
from typing import Sequence

import numpy as np
import numpy.typing as npt
from scipy.ndimage import generic_filter

import doppy
from doppy.product.utils import arr_to_rounded_set


@dataclass
class Options:
    azimuth_offset_deg: float | None


@dataclass
class Wind:
    time: npt.NDArray[np.datetime64]
    height: npt.NDArray[np.float64]
    zonal_wind: npt.NDArray[np.float64]
    meridional_wind: npt.NDArray[np.float64]
    vertical_wind: npt.NDArray[np.float64]
    mask: npt.NDArray[np.bool_]
    system_id: str
    options: Options | None

    @property
    def mask_zonal_wind(self) -> npt.NDArray[np.bool_]:
        return np.isnan(self.zonal_wind)

    @property
    def mask_meridional_wind(self) -> npt.NDArray[np.bool_]:
        return np.isnan(self.meridional_wind)

    @property
    def mask_vertical_wind(self) -> npt.NDArray[np.bool_]:
        return np.isnan(self.vertical_wind)

    @functools.cached_property
    def horizontal_wind_speed(self) -> npt.NDArray[np.float64]:
        return np.sqrt(self.zonal_wind**2 + self.meridional_wind**2)

    @functools.cached_property
    def horizontal_wind_direction(self) -> npt.NDArray[np.float64]:
        direction = np.arctan2(self.zonal_wind, self.meridional_wind)
        direction[direction < 0] += 2 * np.pi
        return np.array(np.degrees(direction), dtype=np.float64)

    @classmethod
    def from_halo_data(
        cls,
        data: Sequence[str]
        | Sequence[Path]
        | Sequence[bytes]
        | Sequence[BufferedIOBase],
        options: Options | None = None,
    ) -> Wind:
        raws = doppy.raw.HaloHpl.from_srcs(data)

        if len(raws) == 0:
            raise doppy.exceptions.NoDataError("HaloHpl data missing")

        raw = (
            doppy.raw.HaloHpl.merge(_select_raws_for_wind(raws))
            .sorted_by_time()
            .non_strictly_increasing_timesteps_removed()
            .nans_removed()
        )
        if len(raw.time) == 0:
            raise doppy.exceptions.NoDataError("No suitable data for the wind product")

        if options and options.azimuth_offset_deg:
            raw.azimuth += options.azimuth_offset_deg

        groups = _group_scans_by_azimuth_rotation(raw)
        time_list = []
        elevation_list = []
        wind_list = []
        rmse_list = []

        for group_index in set(groups):
            pick = group_index == groups
            if pick.sum() < 4:
                continue
            time_, elevation_, wind_, rmse_ = _compute_wind(raw[pick])
            time_list.append(time_)
            elevation_list.append(elevation_)
            wind_list.append(wind_[np.newaxis, :, :])
            rmse_list.append(rmse_[np.newaxis, :])
        time = np.array(time_list)
        if len(time) == 0:
            raise doppy.exceptions.NoDataError(
                "Probably something wrong with scan grouping"
            )
        elevation = np.array(elevation_list)
        wind = np.concatenate(wind_list)
        rmse = np.concatenate(rmse_list)
        if not np.allclose(elevation, elevation[0]):
            raise ValueError("Elevation is expected to stay same")
        height = raw.radial_distance * np.sin(np.deg2rad(elevation[0]))
        mask = _compute_mask(wind, rmse)
        return Wind(
            time=time,
            height=height,
            zonal_wind=wind[:, :, 0],
            meridional_wind=wind[:, :, 1],
            vertical_wind=wind[:, :, 2],
            mask=mask,
            system_id=raw.header.system_id,
            options=options,
        )

    @classmethod
    def from_windcube_data(
        cls,
        data: Sequence[str]
        | Sequence[Path]
        | Sequence[bytes]
        | Sequence[BufferedIOBase],
        options: Options | None = None,
    ) -> Wind:
        raws = doppy.raw.WindCube.from_vad_or_dbs_srcs(data)

        if len(raws) == 0:
            raise doppy.exceptions.NoDataError("WindCube data missing")

        raw = (
            doppy.raw.WindCube.merge(raws)
            .sorted_by_time()
            .non_strictly_increasing_timesteps_removed()
            .reindex_scan_indices()
        )
        # select scans with most frequent elevation angle from range (15,85)
        raw = raw[(raw.elevation > 15) & (raw.elevation < 85)]
        elevation_ints = raw.elevation.round().astype(int)
        unique_elevations, counts = np.unique(elevation_ints, return_counts=True)
        most_frequent_elevation = unique_elevations[np.argmax(counts)]
        raw = raw[elevation_ints == most_frequent_elevation]

        if len(raw.time) == 0:
            raise doppy.exceptions.NoDataError("No suitable data for the wind product")

        if options and options.azimuth_offset_deg:
            raw.azimuth += options.azimuth_offset_deg

        time_list = []
        elevation_list = []
        wind_list = []
        rmse_list = []

        for scan_index in set(raw.scan_index):
            pick = raw.scan_index == scan_index
            if pick.sum() < 4:
                continue
            time_, elevation_, wind_, rmse_ = _compute_wind(raw[pick])
            time_list.append(time_)
            elevation_list.append(elevation_)
            wind_list.append(wind_[np.newaxis, :, :])
            rmse_list.append(rmse_[np.newaxis, :])

        time = np.array(time_list)
        elevation = np.array(elevation_list)
        wind = np.concatenate(wind_list)
        rmse = np.concatenate(rmse_list)
        mask = _compute_mask(wind, rmse) | np.any(np.isnan(wind), axis=2)
        if not np.allclose(elevation, elevation[0]):
            raise ValueError("Elevation is expected to stay same")
        if not (raw.height == raw.height[0]).all():
            raise ValueError("Unexpected heights")
        height = np.array(raw.height[0], dtype=np.float64)
        return Wind(
            time=time,
            height=height,
            zonal_wind=wind[:, :, 0],
            meridional_wind=wind[:, :, 1],
            vertical_wind=wind[:, :, 2],
            mask=mask,
            system_id=raw.system_id,
            options=options,
        )

    @classmethod
    def from_wls70_data(
        cls,
        data: Sequence[str]
        | Sequence[Path]
        | Sequence[bytes]
        | Sequence[BufferedIOBase],
        options: Options | None = None,
    ) -> Wind:
        raws = doppy.raw.Wls70.from_srcs(data)

        if len(raws) == 0:
            raise doppy.exceptions.NoDataError("Wls70 data missing")

        raw = (
            doppy.raw.Wls70.merge(raws)
            .sorted_by_time()
            .non_strictly_increasing_timesteps_removed()
        )

        if options and options.azimuth_offset_deg:
            theta = np.deg2rad(options.azimuth_offset_deg)
            cos_theta = np.cos(theta)
            sin_theta = np.sin(theta)

            meridional_wind = (
                sin_theta * raw.zonal_wind + cos_theta * raw.meridional_wind
            )
            zonal_wind = cos_theta * raw.zonal_wind - sin_theta * raw.meridional_wind
        else:
            meridional_wind = raw.meridional_wind
            zonal_wind = raw.zonal_wind

        mask = (
            np.isnan(raw.meridional_wind)
            | np.isnan(raw.zonal_wind)
            | np.isnan(raw.vertical_wind)
        )
        return Wind(
            time=raw.time,
            height=raw.altitude,
            zonal_wind=zonal_wind,
            meridional_wind=meridional_wind,
            vertical_wind=raw.vertical_wind,
            mask=mask,
            system_id=raw.system_id,
            options=options,
        )

    def write_to_netcdf(self, filename: str | Path) -> None:
        with doppy.netcdf.Dataset(filename) as nc:
            nc.add_dimension("time")
            nc.add_dimension("height")
            nc.add_time(
                name="time",
                dimensions=("time",),
                standard_name="time",
                long_name="Time UTC",
                data=self.time,
                dtype="f8",
            )
            nc.add_variable(
                name="height",
                dimensions=("height",),
                units="m",
                data=self.height,
                dtype="f4",
            )
            nc.add_variable(
                name="uwind_raw",
                dimensions=("time", "height"),
                units="m s-1",
                data=self.zonal_wind,
                mask=self.mask_zonal_wind,
                dtype="f4",
                long_name="Non-screened zonal wind",
            )
            nc.add_variable(
                name="uwind",
                dimensions=("time", "height"),
                units="m s-1",
                data=self.zonal_wind,
                mask=self.mask | self.mask_zonal_wind,
                dtype="f4",
                long_name="Zonal wind",
            )
            nc.add_variable(
                name="vwind_raw",
                dimensions=("time", "height"),
                units="m s-1",
                data=self.meridional_wind,
                mask=self.mask_meridional_wind,
                dtype="f4",
                long_name="Non-screened meridional wind",
            )
            nc.add_variable(
                name="vwind",
                dimensions=("time", "height"),
                units="m s-1",
                data=self.meridional_wind,
                mask=self.mask | self.mask_meridional_wind,
                dtype="f4",
                long_name="Meridional wind",
            )
            nc.add_attribute("serial_number", self.system_id)
            nc.add_attribute("doppy_version", doppy.__version__)
            if self.options is not None and self.options.azimuth_offset_deg is not None:
                nc.add_scalar_variable(
                    name="azimuth_offset",
                    units="degrees",
                    data=self.options.azimuth_offset_deg,
                    dtype="f4",
                    long_name="Azimuth offset of the instrument "
                    "(positive clockwise from north)",
                )


def _compute_wind(
    raw: doppy.raw.HaloHpl | doppy.raw.WindCube,
) -> tuple[float, float, npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    Returns
    -------
    time

    elevation


    wind (range,component):
        Wind components for each range gate.
        Components:
        0: zonal wind
        1: meridional wind
        2: vertical wind

    rmse (range,):
        Root-mean-square error of radial velocity fit for each range gate.

    References
    ----------
    An assessment of the performance of a 1.5 µm Doppler lidar for
    operational vertical wind profiling based on a 1-year trial
        authors: E. Päschke, R. Leinweber, and V. Lehmann
        doi: 10.5194/amt-8-2251-2015
    """
    elevation = np.deg2rad(raw.elevation)
    azimuth = np.deg2rad(raw.azimuth)
    radial_velocity = raw.radial_velocity

    cos_elevation = np.cos(elevation)
    A = np.hstack(
        (
            (np.sin(azimuth) * cos_elevation).reshape(-1, 1),
            (np.cos(azimuth) * cos_elevation).reshape(-1, 1),
            (np.sin(elevation)).reshape(-1, 1),
        )
    )
    A_inv = np.linalg.pinv(A)

    w = A_inv @ radial_velocity
    r_appr = A @ w
    rmse = np.sqrt(np.sum((r_appr - radial_velocity) ** 2, axis=0) / r_appr.shape[0])
    wind = w.T
    time = raw.time[len(raw.time) // 2]
    elevation = np.round(raw.elevation)
    if not np.allclose(elevation, elevation[0]):
        raise ValueError("Elevations in the scan differ")
    return time, elevation[0], wind, rmse


def _compute_mask(
    wind: npt.NDArray[np.float64], rmse: npt.NDArray[np.float64]
) -> npt.NDArray[np.bool_]:
    """
    Parameters
    ----------

    wind (time,range,component)
    intensty (time,range)
    rmse (time,range)
    """

    def neighbour_diff(X: npt.NDArray[np.float64]) -> np.float64:
        mdiff = np.max(np.abs(X - X[len(X) // 2]))
        return np.float64(mdiff)

    WIND_NEIGHBOUR_DIFFERENCE = 20
    neighbour_mask = np.any(
        generic_filter(wind, neighbour_diff, size=(1, 3, 1))
        > WIND_NEIGHBOUR_DIFFERENCE,
        axis=2,
    )

    rmse_th = 5
    return np.array((rmse > rmse_th) | neighbour_mask, dtype=np.bool_)


def _group_scans_by_azimuth_rotation(raw: doppy.raw.HaloHpl) -> npt.NDArray[np.int64]:
    max_timedelta_in_scan = np.timedelta64(30, "s")
    if len(raw.time) < 4:
        raise doppy.exceptions.NoDataError(
            "Less than 4 profiles is not sufficient for wind product."
        )
    groups = -1 * np.ones_like(raw.time, dtype=np.int64)

    group = 0
    first_azimuth_of_scan = _wrap_and_round_angle(raw.azimuth[0])
    groups[0] = group
    for i, (time_prev, time, azimuth) in enumerate(
        zip(raw.time[:-1], raw.time[1:], raw.azimuth[1:]), start=1
    ):
        if (
            angle := _wrap_and_round_angle(azimuth)
        ) == first_azimuth_of_scan or time - time_prev > max_timedelta_in_scan:
            group += 1
            first_azimuth_of_scan = angle
        groups[i] = group
    return groups


def _wrap_and_round_angle(a: np.float64) -> int:
    return int(np.round(a)) % 360


def _select_raws_for_wind(
    raws: Sequence[doppy.raw.HaloHpl],
) -> Sequence[doppy.raw.HaloHpl]:
    counter: Counter[tuple[int, int]] = Counter()
    filtered_raws = []
    for raw in raws:
        select = (1 < raw.elevation) & (raw.elevation < 85)
        for el in arr_to_rounded_set(raw.elevation[select]):
            select_el = raw.elevation.round().astype(int) == el
            select_and = select & select_el

            if _get_nrounded_angles(raw.azimuth[select_and]) > 3:
                filtered_raws.append(raw[select_and])
                counter.update(
                    Counter(
                        (raw.header.mergeable_hash(), el)
                        for el in raw.elevation[select_and].round().astype(int)
                    )
                )
    if len(counter) == 0:
        raise doppy.exceptions.NoDataError(
            "No scans with 1 < elevation angle < 85 and more than 3 azimuth angles"
        )
    if len(counter) == 1:
        return filtered_raws
    # Else select angle closes to 75 from angles
    # that have count larger than mean_count/2
    mean_count = counter.total() / len(counter)
    elevation, _, hash = sorted(
        [
            (el, abs(el - 75), hash)
            for (hash, el), count in counter.items()
            if count > mean_count / 2
        ],
        key=lambda x: x[1],
    )[0]
    elevation_set = {elevation}
    raws = [
        raw
        for raw in filtered_raws
        if raw.header.mergeable_hash() == hash
        and arr_to_rounded_set(raw.elevation) == elevation_set
    ]
    return raws


def _get_nrounded_angles(arr: npt.NDArray[np.float64]) -> int:
    return len(set((x + 360) % 360 for x in arr_to_rounded_set(arr)))



from typing import TypeVar, cast

import numpy as np
from numpy.typing import NDArray

T = TypeVar("T")
NT = TypeVar("NT", bound=np.generic)


def merge_all_equal(key: str, lst: list[T]) -> T:
    if len(set(lst)) != 1:
        raise ValueError(f"Cannot merge header key {key} values {lst}")
    return lst[0]


def merge_all_close(key: str, lst: list[NDArray[NT]]) -> NT:
    if len(lst) == 0:
        raise ValueError(f"Cannot merge empty list for key {key}")
    if any(arr.size == 0 for arr in lst):
        raise ValueError(f"Cannot merge key {key}, one or more arrays are empty.")
    arr = np.concatenate([arr.flatten() for arr in lst])
    if not np.allclose(arr, arr[0]):
        raise ValueError(f"Cannot merge key {key}, values are not close enough")
    return cast(NT, arr[0])



from __future__ import annotations

import io
import re
from dataclasses import dataclass
from datetime import datetime
from io import BufferedIOBase
from pathlib import Path
from typing import Iterable

import numpy as np
import numpy.typing as npt
from numpy import datetime64


@dataclass
class HaloSysParams:
    time: npt.NDArray[datetime64]  # dim: (time, )
    internal_temperature: npt.NDArray[np.float64]  # dim: (time, ), unit: degree Celsius
    internal_relative_humidity: npt.NDArray[np.float64]  # dim: (time, )
    supply_voltage: npt.NDArray[np.float64]  # dim: (time, )
    acquisition_card_temperature: npt.NDArray[np.float64]  # dim: (time, )
    platform_pitch_angle: npt.NDArray[np.float64]  # dim: (time, ), unit: degrees
    platform_roll_angle: npt.NDArray[np.float64]  # dim: (time, ), unit: degrees

    @classmethod
    def from_src(cls, data: str | Path | bytes | BufferedIOBase) -> HaloSysParams:
        if isinstance(data, str):
            path = Path(data)
            with path.open("rb") as f:
                return _from_src(f)
        elif isinstance(data, Path):
            with data.open("rb") as f:
                return _from_src(f)
        elif isinstance(data, bytes):
            return _from_src(io.BytesIO(data))
        elif isinstance(data, BufferedIOBase):
            return _from_src(data)
        else:
            raise TypeError("Unsupported data type")

    @classmethod
    def merge(cls, raws: Iterable[HaloSysParams]) -> HaloSysParams:
        return cls(
            np.concatenate(tuple(r.time for r in raws)),
            np.concatenate(tuple(r.internal_temperature for r in raws)),
            np.concatenate(tuple(r.internal_relative_humidity for r in raws)),
            np.concatenate(tuple(r.supply_voltage for r in raws)),
            np.concatenate(tuple(r.acquisition_card_temperature for r in raws)),
            np.concatenate(tuple(r.platform_pitch_angle for r in raws)),
            np.concatenate(tuple(r.platform_roll_angle for r in raws)),
        )

    def __getitem__(
        self,
        index: int | slice | list[int] | npt.NDArray[np.int64] | npt.NDArray[np.bool_],
    ) -> HaloSysParams:
        if isinstance(index, (int, slice, list, np.ndarray)):
            return HaloSysParams(
                self.time[index],
                self.internal_temperature[index],
                self.internal_relative_humidity[index],
                self.supply_voltage[index],
                self.acquisition_card_temperature[index],
                self.platform_pitch_angle[index],
                self.platform_roll_angle[index],
            )
        raise TypeError

    def sorted_by_time(self) -> HaloSysParams:
        sort_indices = np.argsort(self.time)
        return self[sort_indices]

    def non_strictly_increasing_timesteps_removed(self) -> HaloSysParams:
        is_increasing = np.insert(np.diff(self.time).astype(int) > 0, 0, True)
        return self[is_increasing]


def _correct_concatenated_rows(rows: list[bytes]) -> list[bytes]:
    concat_pattern = re.compile(rb".*(\t[-+0-9]*\.[-+0-9]*\.[-+0-9]*\t).*")

    matches = [concat_pattern.fullmatch(row) for row in rows]

    if not any(matches):
        return rows
    elif not all(matches):
        raise ValueError("Cannot correct the concatenated rows")

    zero_column_pattern = re.compile(rb".*\t0\t.*")
    if not all(zero_column_pattern.fullmatch(row) for row in rows):
        raise ValueError(r"Concatenated rows are expected to have \t0\t pattern")
    rows = [row.replace(b"\t0\t", b"\t") for row in rows]

    new_rows = []
    pattern = re.compile(rb"(.*\t[-+]?[0-9]+\.[0-9]+)([-+][0-9]+\.[0-9]+\t.*)")
    pattern_nan = re.compile(rb"(.*\t)[-+]?[0-9]+\.[0-9]+\.[0-9]+(\t.*)")
    for row in rows:
        m = pattern.fullmatch(row)
        if m:
            new_rows.append(m.group(1) + b"\t" + m.group(2))
        elif m_nan := pattern_nan.fullmatch(row):
            new_rows.append(m_nan.group(1) + b"nan\tnan" + m_nan.group(2))
        else:
            raise ValueError("Cannot separate concatenated floats")
    return new_rows


def _from_src(data: BufferedIOBase) -> HaloSysParams:
    data_bytes = data.read().strip().replace(b",", b".").replace(b"\x00", b"")
    a = data_bytes.strip().split(b"\r\n")
    a = _correct_concatenated_rows(a)
    b = [r.strip().split(b"\t") for r in a]
    arr = np.array(b)
    if arr.shape[1] != 7:
        raise ValueError("Unexpected data format")

    def timestr2datetime64_12H(datetime_bytes: bytes) -> np.datetime64:
        return datetime64(
            datetime.strptime(datetime_bytes.decode("utf-8"), "%m/%d/%Y %I:%M:%S %p"),
            "s",
        )

    def timestr2datetime64_24H(datetime_bytes: bytes) -> np.datetime64:
        return datetime64(
            datetime.strptime(datetime_bytes.decode("utf-8"), "%d/%m/%Y %H:%M:%S"), "s"
        )

    arr_time = np.full_like(arr[:, 0], np.datetime64("NaT"), dtype="datetime64[s]")
    for i, time in enumerate(arr[:, 0]):
        try:
            arr_time[i] = timestr2datetime64_12H(time)
        except ValueError:
            arr_time[i] = timestr2datetime64_24H(time)
    arr_others = np.vectorize(float)(arr[:, 1:])
    return HaloSysParams(arr_time, *arr_others.T)



from collections import Counter

import numpy as np
import numpy.typing as npt


def arr_to_rounded_set(arr: npt.NDArray[np.float64]) -> set[int]:
    return set(int(x) for x in np.round(arr))


def count_rounded(arr: npt.NDArray[np.float64]) -> Counter[int]:
    return Counter(int(x) for x in np.round(arr))



from __future__ import annotations

import io
import re
from dataclasses import dataclass
from datetime import datetime
from io import BufferedIOBase
from pathlib import Path
from typing import Sequence

import numpy as np
import numpy.typing as npt
from numpy import datetime64

from doppy.exceptions import RawParsingError


@dataclass
class HaloBg:
    time: npt.NDArray[datetime64]  # dim: (time, )
    signal: npt.NDArray[np.float64]  # dim: (time, range)

    @property
    def ngates(self) -> int:
        return int(self.signal.shape[1])

    @classmethod
    def from_srcs(
        cls,
        data: Sequence[str]
        | Sequence[Path]
        | Sequence[tuple[bytes, str]]
        | Sequence[tuple[BufferedIOBase, str]],
    ) -> list[HaloBg]:
        """
        Creates a list of `HaloBg` instances from various data sources.

        Parameters
        ----------
        data
            A sequence of data source identifiers which can be file paths (as strings
            or `Path` objects), tuples of raw byte data with filenames, or tuples of
            buffered reader streams with filenames.

        Returns
        -------
        list[HaloBg]
            A list of `HaloBg` instances created from the provided data sources. Data
            sources that cause a raw parsing error are ignored and not included in
            the resulting list.

        Raises
        ------
        TypeError
            If `data` is not a list or tuple of supported types.
        """
        if not isinstance(data, (list, tuple)):
            raise TypeError("data should be list or tuple")
        # TODO: rust reader and proper type checking
        data_normalised = []
        for item in data:
            if isinstance(item, str):
                path = Path(item)
                with path.open("rb") as f:
                    data_normalised.append((f.read(), path.name))
            elif isinstance(item, Path):
                with item.open("rb") as f:
                    data_normalised.append((f.read(), item.name))
            elif isinstance(item, tuple) and isinstance(item[0], bytes):
                data_normalised.append(item)
            elif isinstance(item, tuple) and isinstance(item[0], BufferedIOBase):
                data_normalised.append((item[0].read(), item[1]))
        bgs = []
        for data_bytes, filename in data_normalised:
            try:
                bgs.append(HaloBg.from_src(data_bytes, filename))
            except RawParsingError:
                continue
        return bgs

    @classmethod
    def from_src(
        cls, data: str | Path | bytes | BufferedIOBase, filename: str | None = None
    ) -> HaloBg:
        if isinstance(data, str):
            path = Path(data)
            if filename is None:
                filename = path.name
            with path.open("rb") as f:
                return _from_src(f, filename)
        elif isinstance(data, Path):
            if filename is None:
                filename = data.name
            with data.open("rb") as f:
                return _from_src(f, filename)
        elif isinstance(data, bytes):
            if filename is None:
                raise TypeError("Filename is mandatory if data is given as bytes")
            return _from_src(io.BytesIO(data), filename)
        elif isinstance(data, BufferedIOBase):
            if filename is None:
                raise TypeError(
                    "Filename is mandatory if data is given as BufferedIOBase"
                )
            return _from_src(data, filename)
        else:
            raise TypeError("Unsupported data type")

    @classmethod
    def merge(cls, raws: Sequence[HaloBg]) -> HaloBg:
        return cls(
            np.concatenate(tuple(r.time for r in raws)),
            np.concatenate(tuple(r.signal for r in raws)),
        )

    def __getitem__(
        self,
        index: int
        | slice
        | list[int]
        | npt.NDArray[np.int64]
        | npt.NDArray[np.bool_]
        | tuple[slice, slice],
    ) -> HaloBg:
        if isinstance(index, (int, slice, list, np.ndarray)):
            return HaloBg(self.time[index], self.signal[index])
        elif isinstance(index, tuple):
            return HaloBg(self.time[index[0]], self.signal[index])
        raise TypeError

    def sorted_by_time(self) -> HaloBg:
        sort_indices = np.argsort(self.time)
        return self[sort_indices]

    def non_strictly_increasing_timesteps_removed(self) -> HaloBg:
        is_increasing = np.insert(np.diff(self.time).astype(int) > 0, 0, True)
        return self[is_increasing]


def _from_src(data: BufferedIOBase, filename: str) -> HaloBg:
    if not (m := re.match(r"^Background_(\d{6}-\d{6}).txt", filename)):
        raise ValueError(f"Cannot parse datetime from filename: {filename}")
    time = np.array(datetime64(datetime.strptime(m.group(1), "%d%m%y-%H%M%S")))[
        np.newaxis
    ]

    data_bytes = data.read().strip()
    if b"\r\n" not in data_bytes:
        signal = _from_src_without_newlines(data_bytes)
    else:
        try:
            signal = np.array(list(map(float, data_bytes.split(b"\r\n"))))[np.newaxis]
        except ValueError:
            try:
                signal = np.array(
                    list(map(float, data_bytes.replace(b",", b".").split(b"\r\n")))
                )[np.newaxis]
            except ValueError as err:
                raise RawParsingError(err) from err

    return HaloBg(time, signal)


def _from_src_without_newlines(data: bytes) -> npt.NDArray[np.float64]:
    NUMBER_OF_DECIMALS = 6
    match = re.finditer(rb"\.", data)
    start = 0
    signal = []
    for i in match:
        end = i.end() + NUMBER_OF_DECIMALS
        signal.append(float(data[start:end]))
        start = end
    return np.array(signal)[np.newaxis]



import time


class Timer:
    def __init__(self):
        self.start: float | None = None

    def __enter__(self):
        self.start = time.time()

    def __exit__(self, type, value, traceback):
        if isinstance(self.start, float):
            print(f"Elapsed time: {time.time() - self.start:.2f}")



from io import BufferedIOBase
from pathlib import Path


def bytes_from_src(src: str | bytes | Path | BufferedIOBase) -> bytes:
    if isinstance(src, (str, Path)):
        with open(src, "rb") as f:
            return f.read()
    elif isinstance(src, bytes):
        return src
    elif isinstance(src, BufferedIOBase):
        return src.read()
    else:
        raise TypeError(f"Unexpected type {type(src)} for src")



from enum import Enum


class BgCorrectionMethod(Enum):
    FIT = "fit"
    MEAN = "mean"
    PRE_COMPUTED = "pre_computed"


class BgFitMethod(Enum):
    LIN = "lin"
    EXP = "exp"
    EXPLIN = "explin"



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from io import BufferedIOBase
from pathlib import Path
from typing import DefaultDict, Sequence, Tuple, TypeAlias

import numpy as np
import numpy.typing as npt
import scipy
from scipy.ndimage import median_filter, uniform_filter
from sklearn.cluster import KMeans

import doppy
from doppy import defaults, options
from doppy.product.noise_utils import detect_wind_noise

SelectionGroupKeyType: TypeAlias = tuple[int,]


@dataclass(slots=True)
class RayAccumulationTime:
    # in seconds
    value: float


@dataclass(slots=True)
class PulsesPerRay:
    value: int


@dataclass(slots=True)
class Stare:
    time: npt.NDArray[np.datetime64]
    radial_distance: npt.NDArray[np.float64]
    elevation: npt.NDArray[np.float64]
    beta: npt.NDArray[np.float64]
    snr: npt.NDArray[np.float64]
    radial_velocity: npt.NDArray[np.float64]
    mask_beta: npt.NDArray[np.bool_]
    mask_radial_velocity: npt.NDArray[np.bool_]
    wavelength: float
    system_id: str
    ray_info: RayAccumulationTime | PulsesPerRay

    def __getitem__(
        self,
        index: int
        | slice
        | list[int]
        | npt.NDArray[np.int64]
        | npt.NDArray[np.bool_]
        | tuple[slice, slice],
    ) -> Stare:
        if isinstance(index, (int, slice, list, np.ndarray)):
            return Stare(
                time=self.time[index],
                radial_distance=self.radial_distance,
                elevation=self.elevation[index],
                beta=self.beta[index],
                snr=self.snr[index],
                radial_velocity=self.radial_velocity[index],
                mask_beta=self.mask_beta[index],
                mask_radial_velocity=self.mask_radial_velocity[index],
                wavelength=self.wavelength,
                system_id=self.system_id,
                ray_info=self.ray_info,
            )
        raise TypeError

    @classmethod
    def mask_nan(cls, x: npt.NDArray[np.float64]) -> npt.NDArray[np.bool_]:
        return np.isnan(x)

    @classmethod
    def from_windcube_data(
        cls,
        data: Sequence[str]
        | Sequence[Path]
        | Sequence[bytes]
        | Sequence[BufferedIOBase],
    ) -> Stare:
        raws = doppy.raw.WindCubeFixed.from_srcs(data)
        raw = (
            doppy.raw.WindCubeFixed.merge(raws).sorted_by_time().nan_profiles_removed()
        )

        wavelength = defaults.WindCube.wavelength
        beta = _compute_beta(
            snr=raw.cnr,
            radial_distance=raw.radial_distance,
            wavelength=wavelength,
            beam_energy=defaults.WindCube.beam_energy,
            receiver_bandwidth=defaults.WindCube.receiver_bandwidth,
            focus=defaults.WindCube.focus,
            effective_diameter=defaults.WindCube.effective_diameter,
        )

        mask_beta = _compute_noise_mask_for_windcube(raw)
        mask_radial_velocity = detect_wind_noise(
            raw.radial_velocity, raw.radial_distance, mask_beta
        )

        return cls(
            time=raw.time,
            radial_distance=raw.radial_distance,
            elevation=raw.elevation,
            beta=beta,
            snr=raw.cnr,
            radial_velocity=raw.radial_velocity,
            mask_beta=mask_beta,
            mask_radial_velocity=mask_radial_velocity,
            wavelength=wavelength,
            system_id=raw.system_id,
            ray_info=RayAccumulationTime(raw.ray_accumulation_time.astype(float)),
        )

    @classmethod
    def from_halo_data(
        cls,
        data: Sequence[str]
        | Sequence[Path]
        | Sequence[bytes]
        | Sequence[BufferedIOBase],
        data_bg: Sequence[str]
        | Sequence[Path]
        | Sequence[tuple[bytes, str]]
        | Sequence[tuple[BufferedIOBase, str]],
        bg_correction_method: options.BgCorrectionMethod,
    ) -> Stare:
        raws = doppy.raw.HaloHpl.from_srcs(data)

        if len(raws) == 0:
            raise doppy.exceptions.NoDataError("HaloHpl data missing")

        raw = (
            doppy.raw.HaloHpl.merge(_select_raws_for_stare(raws))
            .sorted_by_time()
            .non_strictly_increasing_timesteps_removed()
            .nans_removed()
        )

        bgs = doppy.raw.HaloBg.from_srcs(data_bg)
        bgs = [bg[:, : raw.header.ngates] for bg in bgs]
        bgs_stare = [bg for bg in bgs if bg.ngates == raw.header.ngates]

        if len(bgs_stare) == 0:
            raise doppy.exceptions.NoDataError("Background data missing")

        bg = (
            doppy.raw.HaloBg.merge(bgs_stare)
            .sorted_by_time()
            .non_strictly_increasing_timesteps_removed()
        )
        raw, intensity_bg_corrected = _correct_background(raw, bg, bg_correction_method)
        if len(raw.time) == 0:
            raise doppy.exceptions.NoDataError("No matching data and bg files")
        intensity_noise_bias_corrected = _correct_intensity_noise_bias(
            raw, intensity_bg_corrected
        )
        wavelength = defaults.Halo.wavelength

        beta = _compute_beta(
            snr=intensity_noise_bias_corrected - 1,
            radial_distance=raw.radial_distance,
            wavelength=wavelength,
            beam_energy=defaults.Halo.beam_energy,
            receiver_bandwidth=defaults.Halo.receiver_bandwidth,
            focus=raw.header.focus_range,
            effective_diameter=defaults.Halo.effective_diameter,
        )

        mask_beta = _compute_noise_mask(
            intensity_noise_bias_corrected, raw.radial_velocity, raw.radial_distance
        )
        mask_radial_velocity = detect_wind_noise(
            raw.radial_velocity, raw.radial_distance, mask_beta
        )

        return cls(
            time=raw.time,
            radial_distance=raw.radial_distance,
            elevation=raw.elevation,
            beta=beta,
            snr=intensity_noise_bias_corrected - 1,
            radial_velocity=raw.radial_velocity,
            mask_beta=mask_beta,
            mask_radial_velocity=mask_radial_velocity,
            wavelength=wavelength,
            system_id=raw.header.system_id,
            ray_info=PulsesPerRay(raw.header.pulses_per_ray),
        )

    def write_to_netcdf(self, filename: str | Path) -> None:
        with doppy.netcdf.Dataset(filename) as nc:
            nc.add_dimension("time")
            nc.add_dimension("range")
            nc.add_time(
                name="time",
                dimensions=("time",),
                standard_name="time",
                long_name="Time UTC",
                data=self.time,
                dtype="f8",
            )
            nc.add_variable(
                name="range",
                dimensions=("range",),
                units="m",
                data=self.radial_distance,
                dtype="f4",
            )
            nc.add_variable(
                name="elevation",
                dimensions=("time",),
                units="degrees",
                data=self.elevation,
                dtype="f4",
                long_name="elevation from horizontal",
            )
            nc.add_variable(
                name="beta_raw",
                dimensions=("time", "range"),
                units="sr-1 m-1",
                data=self.beta,
                dtype="f4",
            )
            nc.add_variable(
                name="beta",
                dimensions=("time", "range"),
                units="sr-1 m-1",
                data=self.beta,
                dtype="f4",
                mask=self.mask_beta,
            )
            nc.add_variable(
                name="v",
                dimensions=("time", "range"),
                units="m s-1",
                long_name="Doppler velocity",
                data=self.radial_velocity,
                dtype="f4",
                mask=self.mask_radial_velocity,
            )
            nc.add_scalar_variable(
                name="wavelength",
                units="m",
                standard_name="radiation_wavelength",
                data=self.wavelength,
                dtype="f4",
            )
            match self.ray_info:
                case RayAccumulationTime(value):
                    nc.add_scalar_variable(
                        name="ray_accumulation_time",
                        units="s",
                        long_name="ray accumulation time",
                        data=value,
                        dtype="f4",
                    )
                case PulsesPerRay(value):
                    nc.add_scalar_variable(
                        name="pulses_per_ray",
                        units="1",
                        long_name="pulses per ray",
                        data=value,
                        dtype="u4",
                    )

            nc.add_attribute("serial_number", self.system_id)
            nc.add_attribute("doppy_version", doppy.__version__)


def _compute_noise_mask_for_windcube(
    raw: doppy.raw.WindCubeFixed,
) -> npt.NDArray[np.bool_]:
    if np.any(np.isnan(raw.cnr)) or np.any(np.isnan(raw.radial_velocity)):
        raise ValueError("Unexpected nans in crn or radial_velocity")

    mask = _mask_with_cnr_norm_dist(raw.cnr) | (np.abs(raw.radial_velocity) > 30)

    cnr = raw.cnr.copy()
    cnr[mask] = np.finfo(float).eps
    cnr_filt = np.array(median_filter(cnr, size=(3, 3)), dtype=np.float64)
    rel_diff = np.abs(cnr - cnr_filt) / np.abs(cnr)
    diff_mask = rel_diff > 0.25

    mask = mask | diff_mask

    return np.array(mask, dtype=np.bool_)


def _mask_with_cnr_norm_dist(cnr: npt.NDArray[np.float64]) -> npt.NDArray[np.bool_]:
    th_trunc = -5.5
    std_factor = 2
    log_cnr = np.log(cnr)
    log_cnr_trunc = log_cnr[log_cnr < th_trunc]
    th_trunc_fit = np.percentile(log_cnr_trunc, 90)
    log_cnr_for_fit = log_cnr_trunc[log_cnr_trunc < th_trunc_fit]
    mean, std = scipy.stats.norm.fit(log_cnr_for_fit)
    return np.array(np.log(cnr) < (mean + std_factor * std), dtype=np.bool_)


def _compute_noise_mask(
    intensity: npt.NDArray[np.float64],
    radial_velocity: npt.NDArray[np.float64],
    radial_distance: npt.NDArray[np.float64],
) -> npt.NDArray[np.bool_]:
    intensity_mean_mask = uniform_filter(intensity, size=(21, 3)) < 1.0025
    velocity_abs_mean_mask = uniform_filter(np.abs(radial_velocity), size=(21, 3)) > 2
    THREE_PULSES_LENGTH = 90
    near_instrument_noise_mask = np.zeros_like(intensity, dtype=np.bool_)
    near_instrument_noise_mask[:, radial_distance < THREE_PULSES_LENGTH] = True
    low_intensity_mask = intensity < 1
    return np.array(
        (intensity_mean_mask & velocity_abs_mean_mask)
        | near_instrument_noise_mask
        | low_intensity_mask,
        dtype=np.bool_,
    )


def _compute_beta(
    snr: npt.NDArray[np.float64],
    radial_distance: npt.NDArray[np.float64],
    wavelength: float,
    beam_energy: float,
    receiver_bandwidth: float,
    focus: float,
    effective_diameter: float,
) -> npt.NDArray[np.float64]:
    """
    Parameters
    ----------
    snr
        for halo: intensity - 1
    radial_distance
        distance from the instrument
    focus
        focal length of the telescope for the transmitter and receiver
    wavelength
        laser wavelength

    Local variables
    ---------------
    eta
        detector quantum efficiency
    E
        beam energy
    nu
        optical frequency
    h
        planc's constant
    c
        speed of light
    B
        receiver bandwidth

    References
    ----------
    Methodology for deriving the telescope focus function and
    its uncertainty for a heterodyne pulsed Doppler lidar
        authors:  Pyry Pentikäinen, Ewan James O'Connor,
            Antti Juhani Manninen, and Pablo Ortiz-Amezcua
        doi: https://doi.org/10.5194/amt-13-2849-2020
    """

    h = scipy.constants.Planck
    c = scipy.constants.speed_of_light
    eta = 1
    E = beam_energy
    B = receiver_bandwidth
    nu = c / wavelength
    A_e = _compute_effective_receiver_energy(
        radial_distance, wavelength, focus, effective_diameter
    )
    beta = 2 * h * nu * B * radial_distance**2 * snr / (eta * c * E * A_e)
    return np.array(beta, dtype=np.float64)


def _compute_effective_receiver_energy(
    radial_distance: npt.NDArray[np.float64],
    wavelength: float,
    focus: float,
    effective_diameter: float,
) -> npt.NDArray[np.float64]:
    """
    NOTE
    ----
    Using uncalibrated values from https://doi.org/10.5194/amt-13-2849-2020


    Parameters
    ----------
    radial_distance
        distance from the instrument
    focus
        effective focal length of the telescope for the transmitter and receiver
    wavelength
        laser wavelength
    """
    D = effective_diameter
    return np.array(
        np.pi
        * D**2
        / (
            4
            * (
                1
                + (np.pi * D**2 / (4 * wavelength * radial_distance)) ** 2
                * (1 - radial_distance / focus) ** 2
            )
        ),
        dtype=np.float64,
    )


def _correct_intensity_noise_bias(
    raw: doppy.raw.HaloHpl, intensity: npt.NDArray[np.float64]
) -> npt.NDArray[np.float64]:
    """
    Parameters
    ----------
    intensity:
        intensity after background correction
    """
    noise_mask = _locate_noise(intensity)
    # Ignore lower gates
    noise_mask[:, raw.radial_distance <= 90] = False

    A_ = np.concatenate(
        (
            raw.radial_distance[:, np.newaxis],
            np.ones((len(raw.radial_distance), 1)),
        ),
        axis=1,
    )[np.newaxis, :, :]
    A = np.tile(
        A_,
        (len(intensity), 1, 1),
    )
    A_noise = np.tile(noise_mask[:, :, np.newaxis], (1, 1, 2))
    A[~A_noise] = 0
    intensity_ = intensity.copy()
    intensity_[~noise_mask] = 0

    A_pinv = np.linalg.pinv(A)
    x = A_pinv @ intensity_[:, :, np.newaxis]
    noise_fit = (A_ @ x).squeeze(axis=2)
    return np.array(intensity / noise_fit, dtype=np.float64)


def _locate_noise(intensity: npt.NDArray[np.float64]) -> npt.NDArray[np.bool_]:
    """
    Returns
    -------
    boolean array M
        where M[i,j] = True if intensity[i,j] contains only noise
        and False otherwise
    """

    INTENSITY_THRESHOLD = 1.008
    MEDIAN_KERNEL_THRESHOLD = 1.002
    GAUSSIAN_THRESHOLD = 0.02

    intensity_normalised = intensity / np.median(intensity, axis=1)[:, np.newaxis]
    intensity_mask = intensity_normalised > INTENSITY_THRESHOLD

    median_mask = (
        scipy.signal.medfilt2d(intensity_normalised, kernel_size=5)
        > MEDIAN_KERNEL_THRESHOLD
    )

    gaussian = scipy.ndimage.gaussian_filter(
        (intensity_mask | median_mask).astype(np.float64), sigma=8, radius=16
    )
    gaussian_mask = gaussian > GAUSSIAN_THRESHOLD

    return np.array(~(intensity_mask | median_mask | gaussian_mask), dtype=np.bool_)


def _correct_background(
    raw: doppy.raw.HaloHpl,
    bg: doppy.raw.HaloBg,
    method: options.BgCorrectionMethod,
) -> Tuple[doppy.raw.HaloHpl, npt.NDArray[np.float64]]:
    """
    Returns
    -------
    raw_with_bg:
        Same as input raw: HaloHpl, but the profiles that does not corresponding
        background measurement have been removed.


    intensity_bg_corrected:
        intensity = SNR + 1 = (A_0 * P_0(z)) / (A_bg * P_bg(z)), z = radial_distance
        The measured background signal P_bg contains usually lots of noise that shows as
        vertical stripes in intensity plots. In bg corrected intensity, P_bg is replaced
        with corrected background profile that should represent the noise floor
        more accurately
    """
    bg_relevant = _select_relevant_background_profiles(bg, raw.time)
    match method:
        case options.BgCorrectionMethod.FIT:
            bg_signal_corrected = _correct_background_by_fitting(
                bg_relevant, raw.radial_distance, fit_method=None
            )
        case options.BgCorrectionMethod.MEAN:
            raise NotImplementedError
        case options.BgCorrectionMethod.PRE_COMPUTED:
            raise NotImplementedError

    raw2bg = np.searchsorted(bg_relevant.time, raw.time, side="right") - 1
    raw_with_bg = raw[raw2bg >= 0]
    raw2bg = raw2bg[raw2bg >= 0]
    raw_bg_original = bg_relevant.signal[raw2bg]
    raw_bg_corrected = bg_signal_corrected[raw2bg]

    intensity_bg_corrected = raw_with_bg.intensity * raw_bg_original / raw_bg_corrected
    return raw_with_bg, intensity_bg_corrected


def _correct_background_by_fitting(
    bg: doppy.raw.HaloBg,
    radial_distance: npt.NDArray[np.float64],
    fit_method: options.BgFitMethod | None,
) -> npt.NDArray[np.float64]:
    clusters = _cluster_background_profiles(bg.signal, radial_distance)
    signal_correcred = np.zeros_like(bg.signal)
    for cluster in set(clusters):
        signal_correcred[clusters == cluster] = _fit_background(
            bg[clusters == cluster], radial_distance, fit_method
        )
    return signal_correcred


def _fit_background(
    bg: doppy.raw.HaloBg,
    radial_distance: npt.NDArray[np.float64],
    fit_method: options.BgFitMethod | None,
) -> npt.NDArray[np.float64]:
    if fit_method is None:
        fit_method = _infer_fit_type(bg.signal, radial_distance)
    match fit_method:
        case options.BgFitMethod.LIN:
            return _linear_fit(bg.signal, radial_distance)
        case options.BgFitMethod.EXP:
            return _exponential_fit(bg.signal, radial_distance)
        case options.BgFitMethod.EXPLIN:
            return _exponential_linear_fit(bg.signal, radial_distance)


def _lin_func(
    x: npt.NDArray[np.float64], radial_distance: npt.NDArray[np.float64]
) -> npt.NDArray[np.float64]:
    return np.array(x[0] * radial_distance + x[1], dtype=np.float64)


def _exp_func(
    x: npt.NDArray[np.float64], radial_distance: npt.NDArray[np.float64]
) -> npt.NDArray[np.float64]:
    return np.array(x[0] * np.exp(x[1] * radial_distance ** x[2]), dtype=np.float64)


def _explin_func(
    x: npt.NDArray[np.float64], radial_distance: npt.NDArray[np.float64]
) -> npt.NDArray[np.float64]:
    return np.array(
        _exp_func(x[:3], radial_distance) + _lin_func(x[3:], radial_distance),
        dtype=np.float64,
    )


def _infer_fit_type(
    bg_signal: npt.NDArray[np.float64], radial_distance: npt.NDArray[np.float64]
) -> options.BgFitMethod:
    peaks = _detect_peaks(bg_signal, radial_distance)
    dist_mask = (90 < radial_distance) & (radial_distance < 8000)
    mask = dist_mask & ~peaks

    scale = np.median(bg_signal, axis=1)[:, np.newaxis]

    rdist = radial_distance[np.newaxis][:, mask]

    signal = (bg_signal / scale)[:, mask]

    def lin_func_rss(x: npt.NDArray[np.float64]) -> np.float64:
        return np.float64(((signal - _lin_func(x, rdist)) ** 2).sum())

    def exp_func_rss(x: npt.NDArray[np.float64]) -> np.float64:
        return np.float64(((signal - _exp_func(x, rdist)) ** 2).sum())

    def explin_func_rss(x: npt.NDArray[np.float64]) -> np.float64:
        return np.float64(((signal - _explin_func(x, rdist)) ** 2).sum())

    method = "Nelder-Mead"
    res_lin = scipy.optimize.minimize(
        lin_func_rss, [1e-5, 1], method=method, options={"maxiter": 2 * 600}
    )
    res_exp = scipy.optimize.minimize(
        exp_func_rss, [1, -1, -1], method=method, options={"maxiter": 3 * 600}
    )
    res_explin = scipy.optimize.minimize(
        explin_func_rss, [1, -1, -1, 0, 0], method=method, options={"maxiter": 5 * 600}
    )

    fit_lin = _lin_func(res_lin.x, rdist)
    fit_exp = _exp_func(res_exp.x, rdist)
    fit_explin = _explin_func(res_explin.x, rdist)

    lin_rss = ((signal - fit_lin) ** 2).sum()
    exp_rss = ((signal - fit_exp) ** 2).sum()
    explin_rss = ((signal - fit_explin) ** 2).sum()

    #
    if exp_rss / lin_rss < 0.95 or explin_rss / lin_rss < 0.95:
        if (exp_rss - explin_rss) / lin_rss > 0.05:
            return options.BgFitMethod.EXPLIN
        else:
            return options.BgFitMethod.EXP
    else:
        return options.BgFitMethod.LIN


def _detect_peaks(
    background_signal: npt.NDArray[np.float64], radial_distance: npt.NDArray[np.float64]
) -> npt.NDArray[np.bool_]:
    """
    background_signal: dim = (time,range)
    radial_distance: dim = (range,)

    Returns a boolean mask, dim = (range, ), where True denotes locations of peaks
    that should be ignored in fitting
    """
    scale = np.median(background_signal, axis=1)[:, np.newaxis]
    bg = background_signal / scale
    return _set_adjacent_true(
        np.concatenate(
            (
                np.array([False]),
                np.diff(np.diff(bg.mean(axis=0))) < -0.01,
                np.array([False]),
            )
        )
    )


def _set_adjacent_true(arr: npt.NDArray[np.bool_]) -> npt.NDArray[np.bool_]:
    temp = np.pad(arr, (1, 1), mode="constant")
    temp[:-2] |= arr
    temp[2:] |= arr
    return temp[1:-1]


def _linear_fit(
    bg_signal: npt.NDArray[np.float64], radial_distance: npt.NDArray[np.float64]
) -> npt.NDArray[np.float64]:
    dist_mask = 90 < radial_distance
    peaks = _detect_peaks(bg_signal, radial_distance)
    mask = dist_mask & ~peaks

    scale = np.median(bg_signal, axis=1)[:, np.newaxis]
    rdist_fit = radial_distance[np.newaxis][:, mask]
    signal_fit = (bg_signal / scale)[:, mask]

    A = np.tile(
        np.concatenate((rdist_fit, np.ones_like(rdist_fit))).T, (signal_fit.shape[0], 1)
    )
    x = np.linalg.pinv(A) @ signal_fit.reshape(-1, 1)
    fit = (
        np.concatenate(
            (radial_distance[:, np.newaxis], np.ones((radial_distance.shape[0], 1))),
            axis=1,
        )
        @ x
    ).T
    return np.array(fit * scale, dtype=np.float64)


def _exponential_fit(
    bg_signal: npt.NDArray[np.float64], radial_distance: npt.NDArray[np.float64]
) -> npt.NDArray[np.float64]:
    dist_mask = 90 < radial_distance
    peaks = _detect_peaks(bg_signal, radial_distance)
    mask = dist_mask & ~peaks
    scale = np.median(bg_signal, axis=1)[:, np.newaxis]
    rdist_fit = radial_distance[np.newaxis][:, mask]
    signal_fit = (bg_signal / scale)[:, mask]

    def exp_func_rss(x: npt.NDArray[np.float64]) -> np.float64:
        return np.float64(((signal_fit - _exp_func(x, rdist_fit)) ** 2).sum())

    result = scipy.optimize.minimize(
        exp_func_rss, [1, -1, -1], method="Nelder-Mead", options={"maxiter": 3 * 600}
    )
    fit = _exp_func(result.x, radial_distance)[np.newaxis, :]
    return np.array(fit * scale, dtype=np.float64)


def _exponential_linear_fit(
    bg_signal: npt.NDArray[np.float64], radial_distance: npt.NDArray[np.float64]
) -> npt.NDArray[np.float64]:
    dist_mask = 90 < radial_distance
    peaks = _detect_peaks(bg_signal, radial_distance)
    mask = dist_mask & ~peaks
    scale = np.median(bg_signal, axis=1)[:, np.newaxis]
    rdist_fit = radial_distance[np.newaxis][:, mask]
    signal_fit = (bg_signal / scale)[:, mask]

    def explin_func_rss(x: npt.NDArray[np.float64]) -> np.float64:
        return np.float64(((signal_fit - _explin_func(x, rdist_fit)) ** 2).sum())

    result = scipy.optimize.minimize(
        explin_func_rss,
        [1, -1, -1, 0, 0],
        method="Nelder-Mead",
        options={"maxiter": 5 * 600},
    )
    fit = _explin_func(result.x, radial_distance)[np.newaxis, :]
    return np.array(fit * scale, dtype=np.float64)


def _select_raws_for_stare(
    raws: Sequence[doppy.raw.HaloHpl],
) -> Sequence[doppy.raw.HaloHpl]:
    if len(raws) == 0:
        raise doppy.exceptions.NoDataError("Expected at least one raw file")
    counter_dd: DefaultDict[tuple[int, int, int], int] = defaultdict(int)
    for raw in raws:
        els, counts = np.unique(np.rint(raw.elevation).astype(int), return_counts=True)
        ngates = len(raw.radial_distance)
        for el, count in zip(els, counts):
            counter_dd[(ngates, el, raw.header.mergeable_hash())] += count
    counter = dict(counter_dd)
    elevation_angle_lb = 75
    elevation_angle_ub = 90
    counter_allowed = {
        (ngates, el, mhash): count
        for (ngates, el, mhash), count in counter.items()
        if (elevation_angle_lb <= el) and (el <= elevation_angle_ub)
    }
    if not counter_allowed:
        raise doppy.exceptions.NoDataError("No raw data suitable for stare product")

    (ngates, elevation, mhash), count = max(counter_allowed.items(), key=lambda x: x[1])
    raws_selected = []
    for raw in raws:
        if len(raw.radial_distance) == ngates and (
            raw.header.mergeable_hash() == mhash
        ):
            select_profiles = np.isclose(raw.elevation, elevation, atol=1)
            raw_selected = raw[select_profiles]
            if raw_selected.time.size != 0:
                raws_selected.append(raw_selected)
    return raws_selected


def _time2bg_time(
    time: npt.NDArray[np.datetime64], bg_time: npt.NDArray[np.datetime64]
) -> npt.NDArray[np.int64]:
    return np.searchsorted(bg_time, time, side="right") - 1


def _select_relevant_background_profiles(
    bg: doppy.raw.HaloBg, time: npt.NDArray[np.datetime64]
) -> doppy.raw.HaloBg:
    """
    expects bg.time to be sorted
    """
    time2bg_time = _time2bg_time(time, bg.time)

    relevant_indices = list(set(time2bg_time[time2bg_time >= 0]))
    bg_ind = np.arange(bg.time.size)
    is_relevant = np.isin(bg_ind, relevant_indices)
    return bg[is_relevant]


def _cluster_background_profiles(
    background_signal: npt.NDArray[np.float64], radial_distance: npt.NDArray[np.float64]
) -> npt.NDArray[np.int64]:
    default_labels = np.zeros(len(background_signal), dtype=int)
    if len(background_signal) < 2:
        return default_labels
    radial_distance_mask = (90 < radial_distance) & (radial_distance < 1500)

    normalised_background_signal = background_signal / np.median(
        background_signal, axis=1, keepdims=True
    )

    profile_median = np.median(
        normalised_background_signal[:, radial_distance_mask], axis=1
    )
    kmeans = KMeans(n_clusters=2, n_init="auto").fit(profile_median[:, np.newaxis])
    cluster_width = np.array([None, None])
    for label in [0, 1]:
        cluster = profile_median[kmeans.labels_ == label]
        cluster_width[label] = np.max(cluster) - np.min(cluster)
    cluster_distance = np.abs(
        kmeans.cluster_centers_[0, 0] - kmeans.cluster_centers_[1, 0]
    )
    max_cluster_width = np.float64(np.max(cluster_width))
    if np.isclose(max_cluster_width, 0):
        return default_labels
    if cluster_distance / max_cluster_width > 3:
        return np.array(kmeans.labels_, dtype=np.int64)
    return default_labels



from __future__ import annotations

from dataclasses import dataclass
from io import BufferedIOBase
from pathlib import Path
from typing import Sequence

import numpy as np
import numpy.typing as npt
from netCDF4 import Dataset, Variable, num2date
from numpy import datetime64

from doppy.utils import merge_all_equal


@dataclass
class WindCubeFixed:
    time: npt.NDArray[datetime64]  # dim: (time, )
    radial_distance: npt.NDArray[np.float64]  # dim: (radial_distance,)
    azimuth: npt.NDArray[np.float64]  # dim: (time, )
    elevation: npt.NDArray[np.float64]  # dim: (time, )
    cnr: npt.NDArray[np.float64]  # dim: (time, radial_distance)
    relative_beta: npt.NDArray[np.float64]  # dim: (time, radial_distance)
    radial_velocity: npt.NDArray[np.float64]  # dim: (time, radial_distance)
    doppler_spectrum_width: npt.NDArray[np.float64]  # dim: (time, radial_distance)
    radial_velocity_confidence: npt.NDArray[np.float64]  # dim: (time, radial_distance)
    ray_accumulation_time: np.float64  # dim: (), unit: seconds
    system_id: str

    @classmethod
    def from_srcs(
        cls,
        data: Sequence[str]
        | Sequence[Path]
        | Sequence[bytes]
        | Sequence[BufferedIOBase],
    ) -> list[WindCubeFixed]:
        return [WindCubeFixed.from_fixed_src(src) for src in data]

    @classmethod
    def from_fixed_src(cls, data: str | Path | bytes | BufferedIOBase) -> WindCubeFixed:
        data_bytes = _src_to_bytes(data)
        nc = Dataset("inmemory.nc", "r", memory=data_bytes)
        return _from_fixed_src(nc)

    @classmethod
    def merge(cls, raws: list[WindCubeFixed]) -> WindCubeFixed:
        return WindCubeFixed(
            time=np.concatenate([r.time for r in raws]),
            radial_distance=_merge_radial_distance_for_fixed(
                [r.radial_distance for r in raws]
            ),
            azimuth=np.concatenate([r.azimuth for r in raws]),
            elevation=np.concatenate([r.elevation for r in raws]),
            radial_velocity=np.concatenate([r.radial_velocity for r in raws]),
            radial_velocity_confidence=np.concatenate(
                [r.radial_velocity_confidence for r in raws]
            ),
            cnr=np.concatenate([r.cnr for r in raws]),
            relative_beta=np.concatenate([r.relative_beta for r in raws]),
            doppler_spectrum_width=np.concatenate(
                [r.doppler_spectrum_width for r in raws]
            ),
            ray_accumulation_time=merge_all_equal(
                "ray_accumulation_time", [r.ray_accumulation_time for r in raws]
            ),
            system_id=merge_all_equal("system_id", [r.system_id for r in raws]),
        )

    def __getitem__(
        self,
        index: int
        | slice
        | list[int]
        | npt.NDArray[np.int64]
        | npt.NDArray[np.bool_]
        | tuple[slice, slice],
    ) -> WindCubeFixed:
        if isinstance(index, (int, slice, list, np.ndarray)):
            return WindCubeFixed(
                time=self.time[index],
                radial_distance=self.radial_distance,
                azimuth=self.azimuth[index],
                elevation=self.elevation[index],
                radial_velocity=self.radial_velocity[index],
                radial_velocity_confidence=self.radial_velocity_confidence[index],
                cnr=self.cnr[index],
                relative_beta=self.relative_beta[index],
                doppler_spectrum_width=self.doppler_spectrum_width[index],
                ray_accumulation_time=self.ray_accumulation_time,
                system_id=self.system_id,
            )
        raise TypeError

    def sorted_by_time(self) -> WindCubeFixed:
        sort_indices = np.argsort(self.time)
        return self[sort_indices]

    def nan_profiles_removed(self) -> WindCubeFixed:
        return self[np.array(~np.all(np.isnan(self.cnr), axis=1), dtype=np.bool_)]


@dataclass
class WindCube:
    time: npt.NDArray[datetime64]  # dim: (time, )
    radial_distance: npt.NDArray[np.float64]  # dim: (time, radial_distance)
    height: npt.NDArray[np.float64]  # dim: (time,radial_distance)
    azimuth: npt.NDArray[np.float64]  # dim: (time, )
    elevation: npt.NDArray[np.float64]  # dim: (time, )
    cnr: npt.NDArray[np.float64]  # dim: (time, radial_distance)
    radial_velocity: npt.NDArray[np.float64]  # dim: (time, radial_distance)
    radial_velocity_confidence: npt.NDArray[np.float64]  # dim: (time, radial_distance)
    scan_index: npt.NDArray[np.int64]
    system_id: str

    @classmethod
    def from_vad_or_dbs_srcs(
        cls,
        data: Sequence[str]
        | Sequence[Path]
        | Sequence[bytes]
        | Sequence[BufferedIOBase],
    ) -> list[WindCube]:
        return [WindCube.from_vad_or_dbs_src(src) for src in data]

    @classmethod
    def from_vad_or_dbs_src(cls, data: str | Path | bytes | BufferedIOBase) -> WindCube:
        data_bytes = _src_to_bytes(data)
        nc = Dataset("inmemory.nc", "r", memory=data_bytes)
        return _from_vad_or_dbs_src(nc)

    @classmethod
    def merge(cls, raws: list[WindCube]) -> WindCube:
        return WindCube(
            scan_index=_merge_scan_index([r.scan_index for r in raws]),
            time=np.concatenate([r.time for r in raws]),
            height=np.concatenate([r.height for r in raws]),
            radial_distance=np.concatenate([r.radial_distance for r in raws]),
            azimuth=np.concatenate([r.azimuth for r in raws]),
            elevation=np.concatenate([r.elevation for r in raws]),
            radial_velocity=np.concatenate([r.radial_velocity for r in raws]),
            radial_velocity_confidence=np.concatenate(
                [r.radial_velocity_confidence for r in raws]
            ),
            cnr=np.concatenate([r.cnr for r in raws]),
            system_id=merge_all_equal("system_id", [r.system_id for r in raws]),
        )

    def __getitem__(
        self,
        index: int
        | slice
        | list[int]
        | npt.NDArray[np.int64]
        | npt.NDArray[np.bool_]
        | tuple[slice, slice],
    ) -> WindCube:
        if isinstance(index, (int, slice, list, np.ndarray)):
            return WindCube(
                time=self.time[index],
                radial_distance=self.radial_distance[index],
                height=self.height[index],
                azimuth=self.azimuth[index],
                elevation=self.elevation[index],
                radial_velocity=self.radial_velocity[index],
                radial_velocity_confidence=self.radial_velocity_confidence[index],
                cnr=self.cnr[index],
                scan_index=self.scan_index[index],
                system_id=self.system_id,
            )
        raise TypeError

    def sorted_by_time(self) -> WindCube:
        sort_indices = np.argsort(self.time)
        return self[sort_indices]

    def non_strictly_increasing_timesteps_removed(self) -> WindCube:
        if len(self.time) == 0:
            return self
        mask = np.ones_like(self.time, dtype=np.bool_)
        latest_time = self.time[0]
        for i, t in enumerate(self.time[1:], start=1):
            if t <= latest_time:
                mask[i] = False
            else:
                latest_time = t
        return self[mask]

    def reindex_scan_indices(self) -> WindCube:
        new_indices = np.zeros_like(self.scan_index)
        indexed = set()
        j = 0
        for i in self.scan_index:
            if i in indexed:
                continue
            new_indices[i == self.scan_index] = j
            indexed.add(i)
            j += 1
        self.scan_index = new_indices
        return self


def _merge_scan_index(index_list: list[npt.NDArray[np.int64]]) -> npt.NDArray[np.int64]:
    if len(index_list) == 0:
        raise ValueError("cannot merge empty list")

    new_index_list = []
    current_max = index_list[0].max()
    new_index_list.append(index_list[0])

    for index_arr in index_list[1:]:
        new_arr = index_arr + current_max + 1
        new_index_list.append(new_arr)
        current_max += index_arr.max() + 1
    return np.concatenate(new_index_list)


def _merge_radial_distance_for_fixed(
    radial_distance_list: list[npt.NDArray[np.float64]],
) -> npt.NDArray[np.float64]:
    if len(radial_distance_list) == 0:
        raise ValueError("cannot merge empty list")
    if not all(
        np.allclose(arr.shape, radial_distance_list[0].shape)
        for arr in radial_distance_list
    ):
        raise ValueError("Cannot merge radial distances with different shapes")
    if not all(
        np.allclose(arr, radial_distance_list[0]) for arr in radial_distance_list
    ):
        raise ValueError("Cannot merge radial distances")
    return radial_distance_list[0]


def _src_to_bytes(data: str | Path | bytes | BufferedIOBase) -> bytes:
    if isinstance(data, str):
        path = Path(data)
        with path.open("rb") as f:
            return f.read()
    elif isinstance(data, Path):
        with data.open("rb") as f:
            return f.read()
    elif isinstance(data, bytes):
        return data
    elif isinstance(data, BufferedIOBase):
        return data.read()


def _from_fixed_src(nc: Dataset) -> WindCubeFixed:
    time_list = []
    cnr_list = []
    relative_beta_list = []
    radial_wind_speed_list = []
    radial_wind_speed_confidence_list = []
    azimuth_list = []
    elevation_list = []
    range_list = []
    doppler_spectrum_width_list = []
    ray_accumulation_time_list = []
    time_reference = (
        nc["time_reference"][:] if "time_reference" in nc.variables else None
    )

    expected_dimensions = ("time", "range")
    for _, group in enumerate(
        nc[group] for group in (nc.variables["sweep_group_name"][:])
    ):
        time_reference_ = time_reference
        if time_reference is None and "time_reference" in group.variables:
            time_reference_ = group["time_reference"][:]

        time_list.append(_extract_datetime64_or_raise(group["time"], time_reference_))
        radial_wind_speed_list.append(
            _extract_float64_or_raise(group["radial_wind_speed"], expected_dimensions)
        )
        cnr_list.append(_extract_float64_or_raise(group["cnr"], expected_dimensions))
        relative_beta_list.append(
            _extract_float64_or_raise(group["relative_beta"], expected_dimensions)
        )
        radial_wind_speed_confidence_list.append(
            _extract_float64_or_raise(
                group["radial_wind_speed_ci"], expected_dimensions
            )
        )
        azimuth_list.append(
            _extract_float64_or_raise(group["azimuth"], expected_dimensions)
        )
        elevation_list.append(
            _extract_float64_or_raise(group["elevation"], expected_dimensions)
        )
        range_list.append(
            _extract_float64_or_raise(group["range"], (expected_dimensions[1],))
        )
        doppler_spectrum_width_list.append(
            _extract_float64_or_raise(
                group["doppler_spectrum_width"], expected_dimensions
            )
        )
        ray_accumulation_time_list.append(
            _extract_float64_or_raise(
                group["ray_accumulation_time"], expected_dimensions
            )
            * 1e-3  # convert ms to s
        )

    return WindCubeFixed(
        time=np.concatenate(time_list),
        radial_distance=np.concatenate(range_list),
        azimuth=np.concatenate(azimuth_list),
        elevation=np.concatenate(elevation_list),
        radial_velocity=np.concatenate(radial_wind_speed_list),
        radial_velocity_confidence=np.concatenate(radial_wind_speed_confidence_list),
        cnr=np.concatenate(cnr_list),
        relative_beta=np.concatenate(relative_beta_list),
        doppler_spectrum_width=np.concatenate(doppler_spectrum_width_list),
        ray_accumulation_time=merge_all_equal(
            "ray_accumulation_time",
            list(np.array(ray_accumulation_time_list, dtype=np.float64)),
        ),
        system_id=nc.instrument_name,
    )


def _from_vad_or_dbs_src(nc: Dataset) -> WindCube:
    scan_index_list: list[npt.NDArray[np.int64]] = []
    time_list: list[npt.NDArray[np.datetime64]] = []
    cnr_list: list[npt.NDArray[np.float64]] = []
    radial_wind_speed_list: list[npt.NDArray[np.float64]] = []
    radial_wind_speed_confidence_list: list[npt.NDArray[np.float64]] = []
    azimuth_list: list[npt.NDArray[np.float64]] = []
    elevation_list: list[npt.NDArray[np.float64]] = []
    range_list: list[npt.NDArray[np.float64]] = []
    height_list: list[npt.NDArray[np.float64]] = []

    time_reference = (
        nc["time_reference"][:] if "time_reference" in nc.variables else None
    )

    expected_dimensions = ("time", "gate_index")
    for i, group in enumerate(
        nc[group] for group in (nc.variables["sweep_group_name"][:])
    ):
        time_reference_ = time_reference
        if time_reference is None and "time_reference" in group.variables:
            time_reference_ = group["time_reference"][:]

        time_list.append(_extract_datetime64_or_raise(group["time"], time_reference_))
        radial_wind_speed_list.append(
            _extract_float64_or_raise(group["radial_wind_speed"], expected_dimensions)
        )
        cnr_list.append(_extract_float64_or_raise(group["cnr"], expected_dimensions))
        radial_wind_speed_confidence_list.append(
            _extract_float64_or_raise(
                group["radial_wind_speed_ci"], expected_dimensions
            )
        )
        azimuth_list.append(
            _extract_float64_or_raise(group["azimuth"], expected_dimensions)
        )
        elevation_list.append(
            _extract_float64_or_raise(group["elevation"], expected_dimensions)
        )
        range_list.append(
            _extract_float64_or_raise(group["range"], expected_dimensions)
        )
        height_list.append(
            _extract_float64_or_raise(group["measurement_height"], expected_dimensions)
        )
        scan_index_list.append(np.full(group["time"][:].shape, i, dtype=np.int64))

    return WindCube(
        scan_index=np.concatenate(scan_index_list),
        time=np.concatenate(time_list),
        radial_distance=np.concatenate(range_list),
        height=np.concatenate(height_list),
        azimuth=np.concatenate(azimuth_list),
        elevation=np.concatenate(elevation_list),
        radial_velocity=np.concatenate(radial_wind_speed_list),
        radial_velocity_confidence=np.concatenate(radial_wind_speed_confidence_list),
        cnr=np.concatenate(cnr_list),
        system_id=nc.instrument_name,
    )


def _extract_datetime64_or_raise(
    nc: Variable[npt.NDArray[np.float64]], time_reference: str | None
) -> npt.NDArray[np.datetime64]:
    match nc.name:
        case "time":
            if nc.dimensions != ("time",):
                raise ValueError

            units = nc.units
            if "time_reference" in nc.units:
                if time_reference is not None:
                    units = nc.units.replace("time_reference", time_reference)
                else:
                    raise ValueError("Unknown time_reference")
            return np.array(num2date(nc[:], units=units), dtype="datetime64[us]")
        case _:
            raise ValueError(f"Unexpected variable name {nc.name}")


def _dB_to_ratio(decibels: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
    return 10 ** (0.1 * decibels)


def _extract_float64_or_raise(
    nc: Variable[npt.NDArray[np.float64]], expected_dimensions: tuple[str, ...]
) -> npt.NDArray[np.float64]:
    match nc.name:
        case "range" | "measurement_height":
            if nc.dimensions != expected_dimensions:
                raise ValueError(f"Unexpected dimensions for {nc.name}")
            if nc.units != "m":
                raise ValueError(f"Unexpected units for {nc.name}")
            if nc[:].mask is not np.bool_(False):
                raise ValueError
            return np.array(nc[:].data, dtype=np.float64)
        case "cnr":
            if nc.dimensions != expected_dimensions:
                raise ValueError(f"Unexpected dimensions for {nc.name}")
            if nc.units != "dB":
                raise ValueError(f"Unexpected units for {nc.name}")
            if nc[:].mask is not np.bool_(False):
                pass  # ignore that array contains masked values
            return _dB_to_ratio(np.array(nc[:].data, dtype=np.float64))
        case "relative_beta":
            if nc.dimensions != expected_dimensions:
                raise ValueError(f"Unexpected dimensions for {nc.name}")
            if nc.units != "m-1 sr-1":
                raise ValueError(f"Unexpected units for {nc.name}")
            if nc[:].mask is not np.bool_(False):
                pass  # ignore that array contains masked values
            return np.array(nc[:].data, dtype=np.float64)
        case "radial_wind_speed":
            if nc.dimensions != expected_dimensions:
                raise ValueError(f"Unexpected dimensions for {nc.name}")
            if nc.units != "m s-1":
                raise ValueError(f"Unexpected units for {nc.name}")
            if nc[:].mask is not np.bool_(False):
                pass  # ignore that array contains masked values
            return np.array(nc[:].data, dtype=np.float64)
        case "radial_wind_speed_ci":
            if nc.dimensions != expected_dimensions:
                raise ValueError(f"Unexpected dimensions for {nc.name}")
            if nc.units != "percent":
                raise ValueError(f"Unexpected units for {nc.name}")
            if nc[:].mask is not np.bool_(False):
                pass  # ignore that array contains masked values
            return np.array(nc[:].data, dtype=np.float64)
        case "azimuth" | "elevation":
            if nc.dimensions != (expected_dimensions[0],):
                raise ValueError(f"Unexpected dimensions for {nc.name}")
            if nc.units != "degrees":
                raise ValueError(f"Unexpected units for {nc.name}")
            if nc[:].mask is not np.bool_(False):
                raise ValueError
            return np.array(nc[:].data, dtype=np.float64)
        case "doppler_spectrum_width":
            if nc.dimensions != expected_dimensions:
                raise ValueError(f"Unexpected dimensions for {nc.name}")
            if nc.units != "m s-1":
                raise ValueError(f"Unexpected units for {nc.name}")
            if nc[:].mask is not np.bool_(False):
                pass  # ignore that array contains masked values
            return np.array(nc[:].data, dtype=np.float64)
        case "ray_accumulation_time":
            if nc.dimensions != ():
                raise ValueError(f"Unexpected dimensions for {nc.name}")
            if nc.units != "ms":
                raise ValueError(f"Unexpected units for {nc.name}")
            if nc[:].mask is not np.bool_(False):
                raise ValueError(f"Variable {nc.name} contains masked values")
            return np.array(nc[:].data, dtype=np.float64)
        case _:
            raise ValueError(f"Unexpected variable name {nc.name}")



import gzip
import logging
import shutil
from io import BytesIO
from pathlib import Path

from requests import Session


def cached_record(
    record: dict, session: Session, check_disk_usage: bool = True
) -> BytesIO:
    cache_dir = Path("cache")
    path = cache_dir / record["uuid"]

    if check_disk_usage:
        HUNDRED_GIGABYTES_AS_BYTES = 100 * 1024 * 1024 * 1024

        _, _, disk_free = shutil.disk_usage("./")
        if disk_free < HUNDRED_GIGABYTES_AS_BYTES:
            _clear_dir(cache_dir)

    if path.is_file():
        return BytesIO(path.read_bytes())
    else:
        path.parent.mkdir(parents=True, exist_ok=True)
        content = session.get(record["downloadUrl"]).content
        if record["filename"].endswith(".gz"):
            try:
                content = gzip.decompress(content)
            except EOFError:
                logging.error(f"Failed to recompress {record['filename']}")
                raise

        with path.open("wb") as f:
            f.write(content)
        return BytesIO(content)


def _clear_dir(path: Path) -> None:
    for f in path.glob("**/*"):
        if f.is_file():
            f.unlink()



from __future__ import annotations

from dataclasses import dataclass

import numpy as np
import numpy.typing as npt
from scipy.interpolate import RegularGridInterpolator


@dataclass
class HorizontalWind:
    time: npt.NDArray[np.datetime64]
    height: npt.NDArray[np.float64]  # Height in meters from reference
    V: npt.NDArray[np.float64]  # Horizontal wind speed in m/s


@dataclass
class VerticalWind:
    time: npt.NDArray[np.datetime64]
    height: npt.NDArray[np.float64]  # Height in meters from reference
    w: npt.NDArray[np.float64]  # Vertical wind speed in m/s
    mask: npt.NDArray[np.bool_]  # mask[t,h] = True iff w[t,h] should be masked


@dataclass
class Options:
    ray_accumulation_time: float  # in seconds
    period: float = 600  # period for computing the variance in seconds
    beam_divergence: float = 33e-6  # radians


@dataclass
class Turbulence:
    time: npt.NDArray[np.datetime64]
    height: npt.NDArray[np.float64]
    turbulent_kinetic_energy_dissipation_rate: npt.NDArray[np.float64]
    mask: npt.NDArray[np.bool_]

    @classmethod
    def from_winds(
        cls, vert: VerticalWind, hori: HorizontalWind, options: Options
    ) -> Turbulence:
        V = _preprocess_horiontal_wind(vert, hori, options)
        ls_low = _length_scale_low(V, vert.height, options)
        res = _compute_variance(vert, options.period)
        sampling_time = _sampling_time_in_seconds(res)
        ls_up = V * sampling_time
        dissipation_rate = _compute_dissipation_rate(res.variance, ls_low, ls_up)
        mask = np.isnan(dissipation_rate) | vert.mask
        return cls(
            time=vert.time.copy(),
            height=vert.height.copy(),
            turbulent_kinetic_energy_dissipation_rate=dissipation_rate,
            mask=mask,
        )


def _sampling_time_in_seconds(r: VarResult) -> npt.NDArray[np.float64]:
    if not all(
        (
            t == np.dtype("datetime64[us]")
            for t in (r.period_start.dtype, r.period_stop.dtype)
        )
    ):
        raise ValueError("period times must be on datetime64[us]")
    td = r.period_stop - r.period_start
    td_in_seconds = td / np.timedelta64(1, "s")
    return np.array(td_in_seconds, dtype=np.float64)


@dataclass
class VarResult:
    variance: npt.NDArray[np.float64]
    period_start: npt.NDArray[np.datetime64]
    period_stop: npt.NDArray[np.datetime64]
    nsamples: npt.NDArray[np.int64]


def _compute_variance(vert: VerticalWind, period: float) -> VarResult:
    # NOTE: numerically unstable

    # To compute actual time window
    next_valid = _next_valid_from_mask(vert.mask)
    prev_valid = _prev_valid_from_mask(vert.mask)

    X = vert.w.copy()
    X[vert.mask] = 0
    X2 = X**2
    X_cumsum = X.cumsum(axis=0)
    X2_cumsum = X2.cumsum(axis=0)

    N_i = (~vert.mask).astype(int)
    N_cumsum = N_i.cumsum(axis=0)

    def N_func(i: int, j: int) -> npt.NDArray[np.float64]:
        return np.array(N_cumsum[j] - N_cumsum[i] + N_i[i], dtype=np.float64)

    def S(i: int, j: int) -> npt.NDArray[np.float64]:
        return np.array(X_cumsum[j] - X_cumsum[i] + X[i], dtype=np.float64)

    def S2(i: int, j: int) -> npt.NDArray[np.float64]:
        return np.array(X2_cumsum[j] - X2_cumsum[i] + X2[i], dtype=np.float64)

    def var_ij(i: int, j: int) -> npt.NDArray[np.float64]:
        N = N_func(i, j)
        with np.errstate(invalid="ignore"):
            return np.array((S2(i, j) - S(i, j) ** 2 / N) / N, dtype=np.float64)

    half_period = np.timedelta64(int(1e6 * period / 2), "us")
    period_start = np.full(vert.w.shape, np.datetime64("NaT", "us"))
    period_stop = np.full(vert.w.shape, np.datetime64("NaT", "us"))
    var = np.full(vert.w.shape, np.nan, dtype=np.float64)
    nsamples = np.zeros_like(vert.w, dtype=np.int64)
    i = 0
    j = 0
    n = len(vert.time)
    for k, t in enumerate(vert.time):
        while i + 1 < n and t - vert.time[i + 1] >= half_period:
            i += 1
        while j + 1 < n and vert.time[j] - t < half_period:
            j += 1
        i_valid = next_valid[i]
        i_inbound = (0 <= i_valid) & (i_valid < n)
        j_valid = prev_valid[j]
        j_inbound = (0 <= j_valid) & (j_valid < n)
        period_start[k][i_inbound] = vert.time[i_valid[i_inbound]]
        period_stop[k][j_inbound] = vert.time[j_valid[j_inbound]]
        var[k] = var_ij(i, j)
        nsamples[k] = N_func(i, j)
    return VarResult(
        variance=var,
        period_start=period_start,
        period_stop=period_stop,
        nsamples=nsamples,
    )


def _length_scale_low(
    V: npt.NDArray[np.float64], height: npt.NDArray[np.float64], opts: Options
) -> npt.NDArray[np.float64]:
    integration_time = opts.ray_accumulation_time
    from_beam = 2 * height * np.sin(opts.beam_divergence / 2)
    from_wind = V * integration_time
    return np.array(from_wind + from_beam[np.newaxis, :], dtype=np.float64)


def _preprocess_horiontal_wind(
    vert: VerticalWind, hori: HorizontalWind, options: Options
) -> npt.NDArray[np.float64]:
    if np.isnan(hori.V).any():
        raise ValueError("horizontal wind speed cannot contains NaNs")
    trg_points = np.meshgrid(vert.time, vert.height, indexing="ij")
    src_points = (hori.time, hori.height)
    src_vals = hori.V

    interp_nearest = RegularGridInterpolator(
        src_points,
        src_vals,
        method="nearest",
        bounds_error=False,
        fill_value=None,
    )
    interp_linear = RegularGridInterpolator(
        src_points, src_vals, method="linear", bounds_error=False
    )
    V_nearest = interp_nearest(trg_points)
    V_linear = interp_linear(trg_points)
    V = V_linear
    V[np.isnan(V)] = V_nearest[np.isnan(V)]
    if np.isnan(V).any():
        raise ValueError("Unexpected NaNs")
    V_rmean = _rolling_mean_over_time(vert.time, V, options.period)
    return V_rmean


def _rolling_mean_over_time(
    time: npt.NDArray[np.datetime64], arr: npt.NDArray[np.float64], period: float
) -> npt.NDArray[np.float64]:
    if arr.ndim != 2:
        raise ValueError("number of dims on arr should be 2")
    if time.ndim != 1 or time.shape[0] != arr.shape[0]:
        raise ValueError("time and arr dimensions do not match")
    if time.dtype != np.dtype("datetime64[us]"):
        raise TypeError(f"Invalid time type: {time.dtype}")

    S = arr.cumsum(axis=0)

    def rolling_mean(i: int, j: int) -> npt.NDArray[np.float64]:
        return np.array((S[j] - S[i] + arr[i]) / (j - i + 1), dtype=np.float64)

    half_period = np.timedelta64(int(period * 0.5e6), "us")
    rol_mean = np.full(arr.shape, np.nan, dtype=np.float64)

    i = 0
    j = 0
    n = len(time)
    for k, t in enumerate(time):
        while i + 1 < n and t - time[i + 1] >= half_period:
            i += 1
        while j + 1 < n and time[j] - t < half_period:
            j += 1
        rol_mean[k] = rolling_mean(i, j)
    return rol_mean


def _compute_dissipation_rate(
    variance: npt.NDArray[np.float64],
    length_scale_lower: npt.NDArray[np.float64],
    length_scale_upper: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Parameters
    ----------
    variance, length_scale_lower, and length_scale_upper
        dimensions: (time,range)
    """
    kolmogorov_constant = 0.55
    with np.errstate(invalid="ignore"):
        dr = (
            2
            * np.pi
            * (2 / (3 * kolmogorov_constant)) ** (3 / 2)
            * variance ** (3 / 2)
            * (length_scale_upper ** (2 / 3) - length_scale_lower ** (2 / 3))
            ** (-3 / 2)
        )
    return np.array(dr, dtype=np.float64)


def _next_valid_from_mask(mask: npt.NDArray[np.bool_]) -> npt.NDArray[np.int64]:
    """
    mask[t,v] (time,value)

    returns N[t,v] = i where i = min { j | j >= t and mask[j,v] == False}
    if the set is non empty and N[t,v] = len(mask) otherwise
    """
    n = len(mask)
    N = np.full(mask.shape, n)
    if mask.size == 0:
        return N
    N[-1][~mask[-1]] = n - 1

    for t in reversed(range(n - 1)):
        N[t][~mask[t]] = t
        N[t][mask[t]] = N[t + 1][mask[t]]
    return np.array(N, dtype=np.int64)


def _prev_valid_from_mask(mask: npt.NDArray[np.bool_]) -> npt.NDArray[np.int64]:
    """
    mask[t,v] (time,value)

    returns N[t,v] = i where i = max { j | j <= t and mask[j,v] == False}
    if the set is non empty and N[t,v] = -1 otherwise
    """
    n = len(mask)
    N = np.full(mask.shape, -1)
    if mask.size == 0:
        return N
    N[0][~mask[0]] = 0
    for t in range(1, n):
        N[t][~mask[t]] = t
        N[t][mask[t]] = N[t - 1][mask[t]]
    return np.array(N, dtype=np.int64)
